#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from typing import Optional, Union

from PyQt5.QtWidgets import QWidget

from kast.Services import Services
from kast.interface.qt5.UiServices import UiServices


class ViewBase:
    def __init__(self, parent: Optional[QWidget]) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self.show()


class ViewModelBase:

    View = Union[ViewBase, QWidget]

    def __init__(self, uiServices: UiServices, view: View) -> None:
        self._uiServices = uiServices
        self.__view = view

    @property
    def services(self) -> Services:
        return self._uiServices.services

    @property
    def uiServices(self) -> UiServices:
        return self._uiServices

    @property
    def view(self) -> View:
        return self.__view



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import os
from typing import Type

from kast.interface.qt5.viewModel.taskbar.TaskbarViewModelBase import TaskbarViewModelBase


def _getImpl() -> Type[TaskbarViewModelBase]:
    if os.name == 'nt':
        from ._win32.Win32TaskbarViewModel import Win32TaskbarViewModel
        return Win32TaskbarViewModel

    if os.name == 'posix':
        from ._unity.UnityTaskbarViewModel import UnityTaskbarViewModel
        return UnityTaskbarViewModel

    return TaskbarViewModelBase


TaskbarViewModel = _getImpl()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from kast.interface.overlay.MediaOverlayBase import MediaOverlayBase
from kast.interface.overlay._winrt.WinrtCastController import WinrtCastController
from kast.media.casting.CastController import CastController
from kast.media.casting.CastState import CastState
from kast.settings.Settings import Settings
from kast.settings.SettingsState import SettingsState


class MediaOverlayWinrt(MediaOverlayBase):

    def __init__(
        self,
        settings: Settings,
        castController: CastController
    ) -> None:
        super().__init__(settings=settings, castController=castController)

        self._winrtCastController = WinrtCastController(settings=settings, castController=castController)

    def run(self) -> None:
        self._winrtCastController.reload()

    def onSettingsEvent(self, event: SettingsState) -> None:
        self._winrtCastController.reload()

    def onCastEvent(self, event: CastState) -> None:
        self._winrtCastController.reload()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from PyQt5.QtGui import QWindow
from PyQt5.QtWidgets import QStyle
from PyQt5.QtWinExtras import QWinThumbnailToolBar, QWinThumbnailToolButton, QWinTaskbarButton

from kast.interface.qt5.UiServices import UiServices
from kast.interface.qt5.service.UiEvent import UiEvent, UiState
from kast.interface.qt5.viewModel.taskbar.TaskbarViewModelBase import TaskbarViewModelBase
from kast.media.casting.CastState import CastMediaState


class Win32TaskbarViewModel(TaskbarViewModelBase):

    def __init__(
        self,
        parent: QWindow,
        uiServices: UiServices
    ) -> None:
        super().__init__(parent, uiServices)
        self.uiServices.uiEventObserver.register(self, self._onUiEvent)

        self._mediaControls = mediaControls = []

        self._taskbarToolbar = taskbarToolbar = QWinThumbnailToolBar(parent)
        taskbarToolbar.setWindow(parent.windowHandle())

        self._taskbarButtonPlayOrPause = buttonPlayOrPause = QWinThumbnailToolButton(taskbarToolbar)
        buttonPlayOrPause.setToolTip('Play/Pause')
        buttonPlayOrPause.setIcon(parent.style().standardIcon(QStyle.SP_MediaPlay))
        mediaControls.append(buttonPlayOrPause)

        self._taskbarButtonStop = buttonStop = QWinThumbnailToolButton(taskbarToolbar)
        buttonStop.setToolTip('Stop')
        buttonStop.setIcon(parent.style().standardIcon(QStyle.SP_MediaStop))
        mediaControls.append(buttonStop)

        self._taskbarButtonSeekForward = buttonSeekForward = QWinThumbnailToolButton(taskbarToolbar)
        buttonSeekForward.setToolTip('Seek forward')
        buttonSeekForward.setIcon(parent.style().standardIcon(QStyle.SP_MediaSeekForward))
        mediaControls.append(buttonSeekForward)

        self._taskbarButtonSeekBackward = buttonSeekBackward = QWinThumbnailToolButton(taskbarToolbar)
        buttonSeekBackward.setToolTip('Seek backward')
        buttonSeekBackward.setIcon(parent.style().standardIcon(QStyle.SP_MediaSeekBackward))
        mediaControls.append(buttonSeekBackward)

        self._taskbarIcon = taskbarIcon = QWinTaskbarButton(parent)
        taskbarIcon.setWindow(parent.windowHandle())
        self._taskbarProgress = taskbarIcon.progress()

        taskbarToolbar.addButton(buttonSeekBackward)
        taskbarToolbar.addButton(buttonPlayOrPause)
        taskbarToolbar.addButton(buttonStop)
        taskbarToolbar.addButton(buttonSeekForward)

        self._enableControls(False)

        buttonPlayOrPause.clicked.connect(self._signalButtonPlayOrPause)
        buttonStop.clicked.connect(self._signalButtonStop)
        buttonSeekForward.clicked.connect(self._signalButtonSeekForward)
        buttonSeekBackward.clicked.connect(self._signalButtonSeekBackward)

    @property
    def _mediaState(self) -> CastMediaState:
        return self.uiServices.uiStateService.castState.mediaState

    def _onUiEvent(self, uiEvent: UiEvent) -> None:
        isStreaming = uiEvent.state == UiState.Streaming
        self._enableControls(isStreaming)

        iconPlayPause = QStyle.SP_MediaPause if self._mediaState.isPlaying() else QStyle.SP_MediaPlay
        self._taskbarButtonPlayOrPause.setIcon(self._parent.style().standardIcon(iconPlayPause))

        self._updateProgress(uiEvent)

    def _signalButtonPlayOrPause(self) -> None:
        self.uiServices.mediaControlService.playOrPause()

    def _signalButtonStop(self) -> None:
        self.uiServices.mediaControlService.stop()

    def _signalButtonSeekForward(self) -> None:
        self.uiServices.mediaControlService.seekForward()

    def _signalButtonSeekBackward(self) -> None:
        self.uiServices.mediaControlService.seekBackward()

    def _enableControls(self, enable: bool = True) -> None:
        for control in self._mediaControls:
            control.setEnabled(enable)

    def _updateProgress(self, uiEvent: UiEvent) -> None:
        if not uiEvent.progress.complete:
            percentage = uiEvent.progress.percentage
            vMax, vCur = (100, percentage) if percentage is not None else (0, 0)
            self._displayProgress(maximum=vMax, value=vCur)
            return

        duration = int(self._mediaState.duration)
        if (uiEvent.state == UiState.Streaming) and (duration > 0):
            self._displayProgress(maximum=duration, value=int(self._mediaState.currentTime))
            return

        self._taskbarProgress.setVisible(False)

    def _displayProgress(self, maximum: int, value: int) -> None:
        self._taskbarProgress.setRange(0, maximum)
        self._taskbarProgress.setValue(value)
        self._taskbarProgress.setVisible(True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import time
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional

from pychromecast.controllers.media import MEDIA_PLAYER_STATE_BUFFERING, MEDIA_PLAYER_STATE_IDLE, \
    MEDIA_PLAYER_STATE_PAUSED, \
    MEDIA_PLAYER_STATE_PLAYING, MEDIA_PLAYER_STATE_UNKNOWN, MediaStatus
from pychromecast.controllers.receiver import CastStatus
from pychromecast.socket_client import CONNECTION_STATUS_CONNECTED, CONNECTION_STATUS_CONNECTING, \
    CONNECTION_STATUS_DISCONNECTED, CONNECTION_STATUS_FAILED, CONNECTION_STATUS_FAILED_RESOLVE, CONNECTION_STATUS_LOST, \
    ConnectionStatus

from kast.utils.chrono import Seconds

DeviceName = str
VolumeLevel = float


def propIgnoreNone(func):
    def wrapper(self, arg):
        if arg is None:
            return

        return func(self, arg)

    return wrapper


class CastConnectionState(Enum):
    Connecting = CONNECTION_STATUS_CONNECTING
    Connected = CONNECTION_STATUS_CONNECTED
    Disconnected = CONNECTION_STATUS_DISCONNECTED
    Failed = CONNECTION_STATUS_FAILED
    FailedResolve = CONNECTION_STATUS_FAILED_RESOLVE
    Lost = CONNECTION_STATUS_LOST


class CastPlayerState(Enum):
    Playing = MEDIA_PLAYER_STATE_PLAYING
    Buffering = MEDIA_PLAYER_STATE_BUFFERING
    Paused = MEDIA_PLAYER_STATE_PAUSED
    Idle = MEDIA_PLAYER_STATE_IDLE
    Unknown = MEDIA_PLAYER_STATE_UNKNOWN


@dataclass
class CastCapabilities:
    canPause: bool = False
    canSeek: bool = False
    canSetMute: bool = False
    canSetVolume: bool = False
    canSkipForward: bool = False
    canSkipBackward: bool = False
    canQueueNext: bool = False
    canQueuePrevious: bool = False


class CastMediaState:
    def __init__(self):
        self._volumeMuted: bool = False
        self._volumeLevel: VolumeLevel = 1.0
        self._playerState: CastPlayerState = CastPlayerState.Unknown
        self._duration: Seconds = Seconds()
        self._currentTime: Seconds = Seconds()
        self._title: str = ''
        self._displayName: str = ''
        self._iconUrl: str = ''
        self._imageUrl: str = ''
        self._contentUrl: str = ''
        self._lastPositionUpdateTimestamp: Optional[Seconds] = None
        self._playInterruptedTimestamp: Optional[Seconds] = None

    @property
    def volumeMuted(self) -> bool:
        return self._volumeMuted

    @volumeMuted.setter
    @propIgnoreNone
    def volumeMuted(self, value: bool) -> None:
        self._volumeMuted = value

    @property
    def volumeLevel(self) -> VolumeLevel:
        return self._volumeLevel

    @volumeLevel.setter
    @propIgnoreNone
    def volumeLevel(self, value: VolumeLevel) -> None:
        self._volumeLevel = value

    @property
    def playerState(self) -> CastPlayerState:
        return self._playerState

    @playerState.setter
    @propIgnoreNone
    def playerState(self, value: CastPlayerState) -> None:
        if self._isPlayerStopping(value):
            self._playInterruptedTimestamp = Seconds(time.time())
        elif self._isPlayerResuming(value):
            self._playInterruptedTimestamp = None

        self._playerState = value

    @property
    def duration(self) -> Seconds:
        return self._duration

    @duration.setter
    @propIgnoreNone
    def duration(self, value: Seconds) -> None:
        self._duration = value

    @property
    def currentTime(self) -> Seconds:
        return self._currentTime + (self._getLatestPlayTimestamp() - self._lastPositionUpdateTimestamp) \
            if self._lastPositionUpdateTimestamp is not None else self._currentTime

    @currentTime.setter
    @propIgnoreNone
    def currentTime(self, value: Seconds) -> None:
        self._currentTime = value
        self._lastPositionUpdateTimestamp = Seconds(time.time())

    @property
    def title(self) -> str:
        return self._title

    @title.setter
    @propIgnoreNone
    def title(self, value: str) -> None:
        self._title = value

    @property
    def displayName(self) -> str:
        return self._displayName

    @displayName.setter
    @propIgnoreNone
    def displayName(self, value: str) -> None:
        self._displayName = value

    @property
    def iconUrl(self) -> str:
        return self._iconUrl

    @iconUrl.setter
    @propIgnoreNone
    def iconUrl(self, value: str) -> None:
        self._iconUrl = value

    @property
    def imageUrl(self) -> str:
        return self._imageUrl

    @imageUrl.setter
    @propIgnoreNone
    def imageUrl(self, value: str) -> None:
        self._imageUrl = value

    @property
    def contentUrl(self) -> str:
        return self._contentUrl

    @contentUrl.setter
    @propIgnoreNone
    def contentUrl(self, value: str) -> None:
        self._contentUrl = value

    def isPlaying(self) -> bool:
        return self.playerState in [CastPlayerState.Playing, CastPlayerState.Buffering]

    def isPaused(self) -> bool:
        return self.playerState == CastPlayerState.Paused

    def isStopped(self) -> bool:
        return not self.isPlaying() and not self.isPaused()

    def _isPlayerStopping(self, newState: CastPlayerState) -> bool:
        return self._playerState == CastPlayerState.Playing and newState != CastPlayerState.Playing

    def _isPlayerResuming(self, newState: CastPlayerState) -> bool:
        return self._playerState != CastPlayerState.Playing and newState == CastPlayerState.Playing

    def _getLatestPlayTimestamp(self) -> Seconds:
        return self._playInterruptedTimestamp if self._playInterruptedTimestamp is not None \
            else Seconds(time.time())


@dataclass
class CastState:
    connection: CastConnectionState = CastConnectionState.Disconnected
    deviceName: Optional[DeviceName] = None
    capabilities: CastCapabilities = CastCapabilities()
    mediaState: CastMediaState = CastMediaState()

    def isConnected(self) -> bool:
        return self.connection in [
            CastConnectionState.Connected,
            CastConnectionState.Connecting
        ]

    def isConnectedOrRecoverable(self) -> bool:
        return self.isConnected() \
           or self.connection == CastConnectionState.Lost

    def update(
        self,
        mediaStatus: Optional[MediaStatus] = None,
        castStatus: Optional[CastStatus] = None,
        connectionStatus: Optional[ConnectionStatus] = None
    ) -> None:
        def validateValue(value: Any, defaultValue: Any) -> Any:
            return defaultValue if value is None else value

        if connectionStatus:
            self.connection = CastConnectionState(validateValue(connectionStatus.status, self.connection.value))

        if castStatus:
            self.mediaState.volumeMuted = castStatus.volume_muted
            self.mediaState.volumeLevel = castStatus.volume_level
            self.mediaState.displayName = castStatus.display_name
            self.mediaState.iconUrl = castStatus.icon_url

        if mediaStatus:
            self.capabilities.canPause = validateValue(mediaStatus.supports_pause, self.capabilities.canPause)
            self.capabilities.canSeek = validateValue(mediaStatus.supports_seek, self.capabilities.canSeek)
            self.capabilities.canSetMute = validateValue(mediaStatus.supports_stream_mute, self.capabilities.canSetMute)
            self.capabilities.canSetVolume = validateValue(mediaStatus.supports_stream_volume, self.capabilities.canSetVolume)
            self.capabilities.canSkipForward = validateValue(mediaStatus.supports_skip_forward, self.capabilities.canSkipForward)
            self.capabilities.canSkipBackward = validateValue(mediaStatus.supports_skip_backward, self.capabilities.canSkipBackward)
            self.capabilities.canQueueNext = validateValue(mediaStatus.supports_queue_next, self.capabilities.canQueueNext)
            self.capabilities.canQueuePrevious = validateValue(mediaStatus.supports_queue_prev, self.capabilities.canQueuePrevious)

            self.mediaState.volumeMuted = mediaStatus.volume_muted
            self.mediaState.volumeLevel = mediaStatus.volume_level
            self.mediaState.playerState = CastPlayerState(mediaStatus.player_state) if mediaStatus.player_state is not None else None
            self.mediaState.duration = Seconds(mediaStatus.duration) if mediaStatus.duration is not None else None
            self.mediaState.currentTime = Seconds(mediaStatus.adjusted_current_time) if mediaStatus.adjusted_current_time is not None else None  # TODO: Maybe we should use not adjusted?
            self.mediaState.title = mediaStatus.title
            self.mediaState.imageUrl = mediaStatus.images[0].url if mediaStatus.images else ''
            self.mediaState.contentUrl = mediaStatus.content_id



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import threading
from pathlib import Path
from typing import Optional

import av

from .MetaData import MetaData
from .SubtitleUtils import SubtitleUtils
from .Transcoder import Codecs, Streams, TranscodeParams, Transcoder


class MediaProcessingService:

    DEFAULT_CONTAINER_FORMAT = 'mp4'

    DEFAULT_CODEC_VIDEO = 'h264'
    DEFAULT_CODEC_AUDIO = 'ac3'

    def __init__(self, storageDir: Path) -> None:
        self._storageDir = storageDir
        self._subtitleUtils = SubtitleUtils(storageDir=self._storageDir)

    @staticmethod
    def extractMetaData(inputFile: Path) -> MetaData:
        def fixStreamName(index: int, name: str, defaultPrefix) -> str:
            return name if name else f'{defaultPrefix} {index}'
        with av.open(str(inputFile)) as f:
            return MetaData(
                title=inputFile.stem,
                audioStreamLangs=[fixStreamName(idx, stream.language, 'Audio') for idx, stream in enumerate(f.streams.audio)],
                subtitleStreamLangs=[fixStreamName(idx, stream.language, 'Subtitles') for idx, stream in enumerate(f.streams.subtitles)]
            )

    def convertSubtitles(self, inputFile: Path) -> Path:
        return self._subtitleUtils.convertToVtt(inputFile=inputFile)

    def extractSubtitles(self, inputFile: Path, streamId: int) -> Path:
        return self._subtitleUtils.extractAsVtt(inputFile=inputFile, streamId=streamId)

    def createTranscoder(
        self,
        inputFile: Path,
        inputStreamIds: Streams,
        outputCodecNames: Optional[Codecs] = None,
        containerFormat: Optional[str] = None,
        progressCallback: Optional[Transcoder.Callback] = None,
        cancelEvent: Optional[threading.Event] = None
    ) -> Transcoder:
        containerFormat = containerFormat if containerFormat else self.DEFAULT_CONTAINER_FORMAT
        outputCodecNames = Codecs(
            video=self.DEFAULT_CODEC_VIDEO,
            audio=self.DEFAULT_CODEC_AUDIO
        )
        return Transcoder(
            params=TranscodeParams(
                inputFile=inputFile,
                inputStreamIds=inputStreamIds,
                outputCodecNames=outputCodecNames
            ),
            outputFile=(self._storageDir / f'movie.{containerFormat}'),
            progressCallback=progressCallback,
            cancelEvent=cancelEvent
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable

from kast.interface.qt5.service.BackgroundRunner import BackgroundRunner
from kast.interface.qt5.service.InterfaceScheduler import InterfaceScheduler

Callback = Callable[[], None]


@dataclass
class Schedulers:
    interfaceScheduler: InterfaceScheduler
    backgroundRunner: BackgroundRunner


class ISchedulable(ABC):
    @property
    @abstractmethod
    def schedulers(self) -> Schedulers:
        """Should return schedulers object."""

    @abstractmethod
    def backgroundTaskMiddleware(self, callback: Callback) -> None:
        """Should directly call callback. Can be used to wrap with try clause to handle errors."""


class SchedulableBase(ISchedulable, ABC):
    def backgroundTaskMiddleware(self, callback: Callback) -> None:
        callback()


class Schedulable(SchedulableBase):

    def __init__(
        self,
        interfaceScheduler: InterfaceScheduler,
        backgroundRunner: BackgroundRunner
    ) -> None:
        self.__schedulers = Schedulers(
            interfaceScheduler,
            backgroundRunner
        )

    @property
    def schedulers(self) -> Schedulers:
        return self.__schedulers



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import itertools
from pathlib import Path
from typing import List, Tuple

import av
import pycaption
import pysubs2


class SubtitleException(Exception):
    pass


class SubtitleUtils:

    _EVENT_LINE_PREFIX = 'Dialogue: '

    def __init__(self, storageDir: Path) -> None:
        self._storageDir = storageDir

        outputFileBaseName = 'subtitles'

        self._intermediateFile = self._storageDir / f'{outputFileBaseName}.ass'
        self._outputFile = self._storageDir / f'{outputFileBaseName}.vtt'

    def extractAsVtt(self, inputFile: Path, streamId: int) -> Path:
        self._extractRawSsa(inputFile=inputFile, streamId=streamId)
        self._convertToVtt(self._convertRawSsaToSrtString())
        return self._outputFile

    def convertToVtt(self, inputFile: Path) -> Path:
        with open(inputFile, encoding='utf-8', errors='ignore') as fInput:
            self._convertToVtt(fInput.read())
        return self._outputFile

    def _convertToVtt(self, subtitles: str) -> None:
        converter = pycaption.CaptionConverter()

        subFormat = pycaption.detect_format(subtitles)
        if not subFormat:
            raise SubtitleException("Could not detect subtitle format or content malformed!")

        converter.read(subtitles, subFormat())
        with open(self._outputFile, 'w', encoding='utf-8', errors='ignore') as fOutput:
            fOutput.write(converter.write(pycaption.WebVTTWriter()))

    def _extractRawSsa(self, inputFile: Path, streamId: int) -> None:
        with av.open(str(inputFile)) as fInput:
            with av.open(str(self._intermediateFile), 'w') as fOutput:
                inStream = fInput.streams.subtitles[streamId]
                outStream = fOutput.add_stream('ass')
                for packet in fInput.demux(inStream):
                    if packet.dts is None:
                        continue
                    packet.stream = outStream
                    fOutput.mux(packet)

    def _convertRawSsaToSrtString(self) -> str:
        subGen = pysubs2.SSAFile()
        with open(self._intermediateFile, encoding='utf-8', errors='ignore') as fInput:
            lineBuffer = ''
            for line in fInput.readlines():
                line = line.rstrip('\n')

                if len(lineBuffer) > 0:
                    if not line.startswith(self._EVENT_LINE_PREFIX):
                        lineBuffer += '\\N'
                    else:
                        subGen.events.append(self._parseSsaEvent(lineBuffer))
                        lineBuffer = ''

                lineBuffer += line

            if len(lineBuffer) > 0:
                subGen.events.append(self._parseSsaEvent(lineBuffer))

        self._removeTextPrefix(subEvents=subGen.events)

        return subGen.to_string('srt')

    def _parseSsaEvent(self, eventStr: str) -> pysubs2.SSAEvent:
        layer, startStr, endStr, text = eventStr[len(self._EVENT_LINE_PREFIX):].split(',', maxsplit=3)

        def timestampStrToMs(timestampStr: str) -> int:
            return pysubs2.time.timestamp_to_ms(pysubs2.time.TIMESTAMP.match(timestampStr).groups())

        event = pysubs2.SSAEvent()
        event.layer = layer
        event.start = timestampStrToMs(startStr)
        event.end = timestampStrToMs(endStr)
        event.text = text

        return event

    def _removeTextPrefix(self, subEvents: List[pysubs2.SSAEvent]) -> None:
        prefix = self._findCommonTextPrefix(subEvents=subEvents)
        if not prefix:
            return

        for subEvent in subEvents:
            subEvent.text = self._removePrefix(subEvent.text, prefix)

    @staticmethod
    def _removePrefix(text: str, prefix: str) -> str:
        if text.startswith(prefix):
            return text[len(prefix):]
        return text

    @staticmethod
    def _findCommonTextPrefix(subEvents: List[pysubs2.SSAEvent]) -> str:
        def allCharacterInColumnAreTheSame(charColumn: Tuple[str]) -> bool:
            return all(charColumn[0] == character for character in charColumn)

        charColumns = zip(*[e.text for e in subEvents])
        prefixColumns = itertools.takewhile(allCharacterInColumnAreTheSame, charColumns)
        prefix = ''.join(prefixColumn[0] for prefixColumn in prefixColumns)

        separator = ','
        return prefix.rsplit(separator, maxsplit=1)[0] + separator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from kast.media.casting.CastController import CastController
from kast.media.casting.CastState import CastState
from kast.settings.Settings import Settings
from kast.settings.SettingsState import SettingsState
from kast.utils.Loggable import Loggable


class MediaOverlayBase(Loggable):

    def __init__(
        self,
        settings: Settings,
        castController: CastController
    ) -> None:
        self._settings = settings
        self._castController = castController

    def run(self) -> None:
        pass

    def onSettingsEvent(self, event: SettingsState) -> None:
        pass

    def onCastEvent(self, event: CastState) -> None:
        pass



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pkg/kast/interface/qt5/view/VideoPreviewView.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_VideoPreviewView(object):
    def setupUi(self, VideoPreviewView):
        VideoPreviewView.setObjectName("VideoPreviewView")
        VideoPreviewView.resize(400, 300)
        self._layout = QtWidgets.QVBoxLayout(VideoPreviewView)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.setObjectName("_layout")
        self._widgetPreview = QtWidgets.QWidget(VideoPreviewView)
        self._widgetPreview.setStyleSheet("background: black")
        self._widgetPreview.setObjectName("_widgetPreview")
        self.layoutPreview = QtWidgets.QVBoxLayout(self._widgetPreview)
        self.layoutPreview.setSpacing(0)
        self.layoutPreview.setObjectName("layoutPreview")
        self.labelImageLogo = QtWidgets.QLabel(self._widgetPreview)
        self.labelImageLogo.setText("")
        self.labelImageLogo.setAlignment(QtCore.Qt.AlignCenter)
        self.labelImageLogo.setObjectName("labelImageLogo")
        self.layoutPreview.addWidget(self.labelImageLogo)
        self._layout.addWidget(self._widgetPreview)

        self.retranslateUi(VideoPreviewView)
        QtCore.QMetaObject.connectSlotsByName(VideoPreviewView)

    def retranslateUi(self, VideoPreviewView):
        _translate = QtCore.QCoreApplication.translate
        VideoPreviewView.setWindowTitle(_translate("VideoPreviewView", "Form"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from mpris_server.server import Server as MprisServer

from kast.interface.overlay.MediaOverlayBase import MediaOverlayBase
from kast.interface.overlay._mpris.MprisCastController import MprisCastController
from kast.interface.overlay._mpris.MprisEventListener import MprisEventListener
from kast.media.casting.CastController import CastController
from kast.media.casting.CastState import CastState
from kast.settings.Settings import Settings
from kast.settings.SettingsState import SettingsState


class MediaOverlayMpris(MediaOverlayBase):

    def __init__(
        self,
        settings: Settings,
        castController: CastController
    ) -> None:
        super().__init__(settings=settings, castController=castController)

        self._mprisCastController = MprisCastController(settings=settings, castController=castController)
        self._mprisServer = MprisServer(name=settings.appName, adapter=self._mprisCastController)
        self._mprisServer.publish()
        self._mprisEventListener = MprisEventListener(mprisServer=self._mprisServer, mprisCastController=self._mprisCastController)

    def run(self) -> None:
        self._mprisServer.loop()

    def onSettingsEvent(self, event: SettingsState) -> None:
        self._mprisEventListener.onSettingsEvent(event=event)

    def onCastEvent(self, event: CastState) -> None:
        self._mprisEventListener.onCastEvent(event=event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from typing import Callable
from queue import Empty, Full, Queue
from typing import Optional

from PyQt5.QtCore import QTimer

from kast.utils.Loggable import Loggable


class InterfaceScheduler(Loggable):

    Callback = Callable[[], None]

    TIMEOUT_SECONDS = 0.1

    def __init__(self) -> None:
        self._queue = Queue()
        self._timer = QTimer()
        self._timer.setInterval(100)
        self._timer.timeout.connect(self._process)
        self._timer.start()

    def schedule(self, callback: Callback) -> None:
        try:
            self._queue.put(item=callback, block=True, timeout=self.TIMEOUT_SECONDS)
        except Full:
            raise RuntimeError(f"Could not schedule interface task! Queue size: {self._queue.qsize()}")

    def _process(self) -> None:
        callback = self._tryPopTask()
        if callback:
            callback()
            self._queue.task_done()

    def _tryPopTask(self) -> Optional[Callback]:
        try:
            return self._queue.get(block=True, timeout=self.TIMEOUT_SECONDS)
        except Empty:
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import shutil
import tempfile
import threading
from pathlib import Path

from kast import __app_name__, __version__, __author__, __email__, __package_name__
from kast.settings.SettingsEventObserver import SettingsEventObserver
from kast.settings.SettingsState import SettingsState
from kast.utils.ResourceProvider import ResourceProvider


class Settings:

    def __init__(
        self,
        resourceProvider: ResourceProvider,
        settingsEventObserver: SettingsEventObserver
    ) -> None:
        self._resourceProvider = resourceProvider
        self._settingsEventObserver = settingsEventObserver

        self._persistentStoragePath = Path.home().joinpath(f".config/{__package_name__}")
        self._persistentStoragePath.mkdir(parents=True, exist_ok=True)
        self._settingsStateFilePath = self._persistentStoragePath.joinpath('settings.json')
        self._temporaryStorage = tempfile.TemporaryDirectory(prefix=f"{__package_name__}-")

        self._lock = threading.Lock()
        self._settingsState = SettingsState()

        self._load()
        self._cleanupArtifacts()

    @property
    def author(self) -> str:
        return __author__

    @property
    def email(self) -> str:
        return __email__

    @property
    def package(self) -> str:
        return __package_name__

    @property
    def appName(self) -> str:
        return __app_name__

    @property
    def appVersion(self) -> str:
        return __version__

    @property
    def desktopFileName(self) -> str:
        return f'{__package_name__}.desktop'

    @property
    def persistentStoragePath(self) -> Path:
        return self._persistentStoragePath

    @property
    def temporaryStoragePath(self) -> Path:
        return Path(self._temporaryStorage.name)

    @property
    def logFilePath(self) -> Path:
        return self.persistentStoragePath.joinpath(f"{self.package}.log")

    @property
    def appIconPath(self) -> Path:
        return self._resourceProvider.getResourcePath(Path('appicon.png'))

    @property
    def browseMediaDir(self) -> Path:
        with self._lock:
            return Path(self._settingsState.browseMediaPath)

    @browseMediaDir.setter
    def browseMediaDir(self, value: Path) -> None:
        with self._lock:
            self._settingsState.browseMediaPath = str(value)
        self._notify()

    def _notify(self) -> None:
        self._settingsEventObserver.notify(event=self._settingsState)
        self._save()

    def _cleanupArtifacts(self) -> None:
        if self._settingsState.previousTemporaryPath and Path(self._settingsState.previousTemporaryPath).exists():
            shutil.rmtree(self._settingsState.previousTemporaryPath)

        self._settingsState.previousTemporaryPath = str(self.temporaryStoragePath)

    def _load(self) -> None:
        with self._lock:
            if self._settingsStateFilePath.is_file() and self._settingsStateFilePath.stat().st_size:
                with self._settingsStateFilePath.open() as fInput:
                    self._settingsState = SettingsState.from_json(fInput.read())

    def _save(self) -> None:
        with self._lock:
            with self._settingsStateFilePath.open(mode='w') as fOutput:
                fOutput.write(self._settingsState.to_json())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from kast.Services import Services
from kast.interface.qt5.service.BackgroundRunner import BackgroundRunner
from kast.interface.qt5.service.InterfaceScheduler import InterfaceScheduler
from kast.interface.qt5.service.MediaControlService import MediaControlService
from kast.interface.qt5.service.UiEventObserver import UiEventObserver
from kast.interface.qt5.service.UiStateService import UiStateService


class UiServices:

    def __init__(self, services: Services) -> None:
        self.services = services
        self.interfaceScheduler = InterfaceScheduler()
        self.backgroundRunner = BackgroundRunner(interfaceScheduler=self.interfaceScheduler)
        self.uiEventObserver = UiEventObserver()
        self.uiStateService = UiStateService(
            interfaceScheduler=self.interfaceScheduler,
            uiEventObserver=self.uiEventObserver,
            castEventObserver=self.services.castEventObserver
        )
        self.mediaControlService = MediaControlService(
            services=services,
            interfaceScheduler=self.interfaceScheduler,
            backgroundRunner=self.backgroundRunner,
            uiStateService=self.uiStateService
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from PyQt5.QtWidgets import QMainWindow

from kast.interface.qt5.UiServices import UiServices
from kast.interface.qt5.view.MainWindowView import Ui_MainWindowView
from kast.interface.qt5.viewModel.AboutViewModel import AboutViewModel
from kast.interface.qt5.viewModel.MediaControlViewModel import MediaControlViewModel
from kast.interface.qt5.viewModel.StatusBarViewModel import StatusBarViewModel
from kast.interface.qt5.viewModel.VideoPreviewViewModel import VideoPreviewViewModel
from kast.interface.qt5.viewModel.VideoSettingsViewModel import VideoSettingsViewModel
from kast.interface.qt5.viewModel.ViewModelBase import ViewBase, ViewModelBase
from kast.interface.qt5.viewModel.taskbar.TaskbarViewModel import TaskbarViewModel


class View(ViewBase, QMainWindow, Ui_MainWindowView):
    pass


class MainWindowViewModel(ViewModelBase):

    def __init__(self, uiServices: UiServices) -> None:
        self._view = View(parent=None)
        super().__init__(uiServices=uiServices, view=self._view)
        self._view.setWindowTitle(self.services.settings.appName)

        self._taskbarViewModel = TaskbarViewModel(parent=self._view, uiServices=uiServices)

        self._aboutViewModel = AboutViewModel(parent=self._view, uiServices=self.uiServices)
        self._videoPreviewViewModel = VideoPreviewViewModel(parent=self.view, uiServices=self.uiServices)
        self._mediaControlViewModel = MediaControlViewModel(parent=self.view, uiServices=self.uiServices, videoPreviewViewModel=self._videoPreviewViewModel)
        self._videoSettingsViewModel = VideoSettingsViewModel(parent=self.view, uiServices=self.uiServices)
        self._statusBarViewModel = StatusBarViewModel(parent=self.view, uiServices=self.uiServices)

        self._view.layoutForPreview.addWidget(self._videoPreviewViewModel.view)
        self._addControlsViewModel(self._mediaControlViewModel)
        self._addControlsViewModel(self._videoSettingsViewModel)
        self._addControlsViewModel(self._statusBarViewModel)

        self._view.actionExit.triggered.connect(self._view.close)
        self._view.actionAbout.triggered.connect(self._aboutViewModel.view.show)

    def _addControlsViewModel(self, viewModel: ViewModelBase) -> None:
        self._view.layoutForControls.addWidget(viewModel.view)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import math
import threading
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, List, Optional

import av

from kast.media.processing.common import CodecName, StreamId, containerExtension
from kast.utils.Loggable import Loggable
from kast.media.processing.Resolution import FULL_HD, Resolution


@dataclass(frozen=True)
class Streams:
    video: StreamId = 0
    audio: StreamId = 0


@dataclass(frozen=True)
class Codecs:
    video: CodecName
    audio: CodecName


@dataclass(frozen=True)
class TranscodeParams:
    inputFile: Path
    inputStreamIds: Streams
    outputCodecNames: Codecs
    maxResolution: Resolution = FULL_HD


class Transcoder(Loggable):

    Callback = Callable[[int, bool], None]

    PROGRESS_UPDATE_DELAY = 1

    def __init__(
        self,
        params: TranscodeParams,
        outputFile: Path,
        progressCallback: Optional[Callback] = None,
        cancelEvent: Optional[threading.Event] = None
    ) -> None:
        self._params = params
        self._outputFile = outputFile
        self._progressCallback = progressCallback

        self._inputCodeNames = self._extractInputCodecs(
            inputFile=self._params.inputFile,
            inputStreams=self._params.inputStreamIds
        )

        self._inputResolution = self._extractInputResolution(
            inputFile=self._params.inputFile,
            videoStreamId=self._params.inputStreamIds.video
        )
        self._outputResolution = self._inputResolution.shrinkToFit(self._params.maxResolution)

        self._remuxContainer = (containerExtension(outputFile) not in self._getFormatExtensions(self._params.inputFile))
        self._transcodeAudio = (self._params.outputCodecNames.audio != self._inputCodeNames.audio)
        self._transcodeVideo = (self._params.outputCodecNames.video != self._inputCodeNames.video) or \
           (self._inputResolution > self._params.maxResolution)

        self._cancelEvent = cancelEvent if cancelEvent else threading.Event()
        self._complete = False
        self._lastPos = 0

    @property
    def requireProcessing(self) -> bool:
        return self._remuxContainer or self._transcodeAudio or self._transcodeVideo

    @property
    def inputFile(self) -> Path:
        return self._params.inputFile

    @property
    def outputFile(self) -> Path:
        return self._outputFile if self.requireProcessing else self._params.inputFile

    @property
    def complete(self) -> bool:
        return self._complete

    @property
    def lastPos(self) -> int:
        return self._lastPos

    @property
    def inputCodecNames(self) -> Codecs:
        return self._inputCodeNames

    @property
    def outputCodecNames(self) -> Codecs:
        return self._params.outputCodecNames

    @property
    def maxResolution(self) -> Resolution:
        return self._params.maxResolution

    @property
    def inputResolution(self) -> Resolution:
        return self._inputResolution

    @property
    def outputResolution(self) -> Resolution:
        return self._outputResolution

    @property
    def params(self) -> TranscodeParams:
        return self._params

    @property
    def cancelEvent(self) -> threading.Event:
        return self._cancelEvent

    def run(self) -> bool:
        if self._complete:
            return self._complete
        if not self.requireProcessing:
            self._complete = True
            self._notifyProgress(100)
            return self._complete

        with av.open(str(self._params.inputFile)) as fInput,\
                av.open(str(self._outputFile), 'w') as fOutput:

            inStreamVideo = fInput.streams.video[self._params.inputStreamIds.video]
            inStreamAudio = fInput.streams.audio[self._params.inputStreamIds.audio]

            self.log.info(f"Remuxing: {containerExtension(self.inputFile)} -> {containerExtension(self.outputFile)}")
            self.log.info(f"Transcoding: {self.inputCodecNames} -> {self.outputCodecNames}")
            self.log.info(f"Resizing: {self.inputResolution} -> {self.outputResolution}")

            def createOutputVideoStream() -> av.video.stream.VideoStream:
                if not self._transcodeVideo:
                    return fOutput.add_stream(template=inStreamVideo)
                vStream = fOutput.add_stream(self._params.outputCodecNames.video, inStreamVideo.average_rate)
                vStream.width = self.outputResolution.width
                vStream.height = self.outputResolution.height
                vStream.options = {'crf': '23'}
                return vStream

            def createOutputAudioStream() -> av.audio.stream.AudioStream:
                if not self._transcodeAudio:
                    return fOutput.add_stream(template=inStreamAudio)
                return fOutput.add_stream(self._params.outputCodecNames.audio)

            outStreamVideo = createOutputVideoStream()
            outStreamAudio = createOutputAudioStream()

            last_update = 0
            for packet in fInput.demux([inStreamAudio, inStreamVideo]):
                if self._cancelEvent.is_set():
                    return False

                if packet is not None and packet.pos is not None:
                    self._lastPos = packet.pos
                    if (time.time() - last_update) > self.PROGRESS_UPDATE_DELAY:
                        self._notifyProgress(math.floor((packet.pos/fInput.size)*100))
                        last_update = time.time()

                if packet.stream.type == 'video':
                    self._processPacket(packet, outStreamVideo, fOutput, self._transcodeVideo)
                else:
                    self._processPacket(packet, outStreamAudio, fOutput, self._transcodeAudio)

            self._flushStream(outStreamVideo, fOutput, self._transcodeVideo)
            self._flushStream(outStreamAudio, fOutput, self._transcodeAudio)

        self._complete = True
        self._notifyProgress(100)

        return self._complete

    def _notifyProgress(self, progress: int) -> None:
        if self._progressCallback:
            self._progressCallback(progress, self._complete)

    def _processPacket(
        self,
        packet: av.packet.Packet,
        stream: av.stream.Stream,
        outContainer: av.container.OutputContainer,
        doTranscode: bool
    ) -> None:
        if doTranscode:
            self._transcode(packet, stream, outContainer)
            return
        self._remux(packet, stream, outContainer)

    @staticmethod
    def _transcode(
        packet: av.packet.Packet,
        stream: av.stream.Stream,
        outContainer: av.container.OutputContainer
    ) -> None:
        for frame in packet.decode():
            frame.pts = None
            frame.time_base = None
            outContainer.mux(stream.encode(frame))

    @staticmethod
    def _remux(
        packet: av.packet.Packet,
        stream: av.stream.Stream,
        outContainer: av.container.OutputContainer
    ) -> None:
        packet.stream = stream
        if packet.dts:
            outContainer.mux(packet)

    @staticmethod
    def _flushStream(
        stream: av.stream.Stream,
        outContainer: av.container.OutputContainer,
        doTranscode: bool
    ) -> None:
        if doTranscode:
            outContainer.mux(stream.encode())

    @staticmethod
    def _getFormatExtensions(filePath: Path) -> List[str]:
        with av.open(str(filePath)) as fInput:
            return fInput.format.extensions

    @staticmethod
    def _extractInputResolution(inputFile: Path, videoStreamId: StreamId) -> Resolution:
        with av.open(str(inputFile)) as fInput:
            stream = fInput.streams.video[videoStreamId]
            return Resolution(width=stream.width, height=stream.height)

    @staticmethod
    def _extractInputCodecs(inputFile: Path, inputStreams: Streams) -> Codecs:
        with av.open(str(inputFile)) as fInput:
            return Codecs(
                audio=fInput.streams.audio[inputStreams.audio].codec_context.name,
                video=fInput.streams.video[inputStreams.video].codec_context.name
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from typing import List, Optional

import pychromecast
from pychromecast.controllers.media import MediaStatus, MediaStatusListener
from pychromecast.controllers.receiver import CastStatus, CastStatusListener
from pychromecast.socket_client import ConnectionStatus, ConnectionStatusListener

from kast.media.processing.common import CodecName
from kast.utils.Loggable import Loggable
from kast.media.casting.CastEventObserver import CastEventObserver
from kast.media.casting.CastException import CastException
from kast.media.casting.CastState import CastMediaState, CastState, Seconds, \
    VolumeLevel, DeviceName
from kast.settings.Settings import Settings


class CastController(
    Loggable,
    MediaStatusListener,
    CastStatusListener,
    ConnectionStatusListener
):
    TIMEOUT = Seconds(10)
    SEEK_DELTA = Seconds(10)

    def __init__(
        self,
        settings: Settings,
        castEventObserver: CastEventObserver
    ) -> None:
        self._settings = settings
        self._castEventObserver = castEventObserver

        self.__device: Optional[pychromecast.Chromecast] = None
        self._castState: CastState = CastState()

    def _clear(self, mediaOnly: bool = False) -> None:
        if mediaOnly:
            self._castState.mediaState = CastMediaState()
            return

        self._castState = CastState()

    @property
    def _device(self) -> Optional[pychromecast.Chromecast]:
        return self.__device if self.isConnected else None

    @property
    def _mediaController(self) -> Optional[pychromecast.controllers.media.MediaController]:
        device = self._device
        return device.media_controller if device else None

    @property
    def castState(self) -> CastState:
        return self._castState

    @property
    def isConnected(self):
        return self.__device and self.castState.isConnected()

    @property
    def supportedMimeTypes(self) -> List[str]:
        return ['video/mp4']  # Theoretically should also support: ['video/x-matroska', 'video/webm']

    @property
    def supportedContainerFormats(self) -> List[str]:
        return ['mp4']  # Theoretically should also support: ['mkv', 'webm']

    @property
    def supportedVideoCodecs(self) -> List[CodecName]:
        return ['h264']  # Since 'Chromecast Ultra' support for 'hevc' (h265)!

    @property
    def supportedAudioCodecs(self) -> List[CodecName]:
        return ['ac3']  # No evidence for wider support for 'aac'.

    @property
    def preferredContainerFormat(self) -> str:  # TODO:  Apply some logic if possible.
        return self.supportedContainerFormats[0]

    @property
    def preferredVideoCodec(self) -> CodecName:  # TODO:  Apply some logic if possible.
        return self.supportedVideoCodecs[0]

    @property
    def preferredAudioCodec(self) -> CodecName:  # TODO:  Apply some logic if possible.
        return self.supportedAudioCodecs[0]

    def searchDevices(self) -> List[DeviceName]:
        castInfoList, castBrowser = pychromecast.discovery.discover_chromecasts(timeout=self.TIMEOUT.value)
        castBrowser.stop_discovery()
        return [ci.friendly_name for ci in castInfoList]

    def connect(self, name: DeviceName) -> None:
        chromecasts, castBrowser = pychromecast.get_chromecasts(timeout=self.TIMEOUT.value)
        castBrowser.stop_discovery()
        for chromecast in chromecasts:
            if name == chromecast.name:
                self._setupDevice(chromecast)
                return
        raise CastException(f"Could not find device by name '{name}'!")

    def disconnect(self) -> None:
        device = self._device
        if device:
            device.quit_app()
            device.disconnect()
            self._clear()

    def stream(
        self,
        movieUrl: str,
        movieMime: str = 'video/mp4',
        subtitlesUrl: str = None,
        subtitlesMime: str = 'text/vtt',
        thumbnailUrl: str = None,
        play: bool = True,
        title: str = None
    ) -> None:
        title = title if title is not None else self._settings.appName
        mediaController = self._mediaController
        if mediaController:
            mediaController.play_media(
                url=movieUrl,
                content_type=movieMime,
                subtitles=subtitlesUrl,
                subtitles_mime=subtitlesMime,
                autoplay=play,
                title=title,
                thumb=thumbnailUrl
            )
            mediaController.block_until_active()

    def quit(self) -> None:
        device = self._device
        device and device.quit_app()

    def setMute(self, val: bool) -> None:
        device = self._device
        device and device.set_volume_muted(val)

    def setVolume(self, val: VolumeLevel) -> None:
        device = self._device
        device and device.set_volume(val)

    def play(self) -> None:
        mediaController = self._mediaController
        mediaController and mediaController.play()

    def pause(self) -> None:
        mediaController = self._mediaController
        mediaController and mediaController.pause()

    def resume(self) -> None:
        mediaController = self._mediaController
        mediaController and mediaController.play()

    def stop(self) -> None:
        mediaController = self._mediaController
        mediaController and mediaController.stop()

    def seek(self, timePos: Seconds) -> None:
        mediaController = self._mediaController
        mediaController and mediaController.seek(timePos.value)

    def seekForward(self) -> None:
        self.seek(self._castState.mediaState.currentTime + self.SEEK_DELTA)

    def seekBackward(self) -> None:
        self.seek(self._castState.mediaState.currentTime - self.SEEK_DELTA)

    def _setupDevice(self, device: pychromecast.Chromecast) -> None:
        if device.cast_type != pychromecast.CAST_TYPE_CHROMECAST:
            raise CastException(f"Device '{device.name}' does not support video casting!")

        device.register_status_listener(self)
        device.media_controller.register_status_listener(self)
        device.socket_client.register_connection_listener(self)

        device.wait()

        self.__device = device
        self._castState.deviceName = device.name

    def _notify(self) -> None:
        self._castEventObserver.notify(event=self._castState)

    def new_media_status(self, status: MediaStatus):
        self._castState.update(mediaStatus=status)

        if self._castState.mediaState.isStopped():
            self._clear(mediaOnly=True)

        self._notify()

    def new_cast_status(self, status: CastStatus):
        self._castState.update(castStatus=status)
        self._notify()

    def new_connection_status(self, status: ConnectionStatus):
        self._castState.update(connectionStatus=status)

        if not self._castState.isConnectedOrRecoverable():
            self._clear()

        self._notify()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import sys
from typing import Callable

from PyQt5.QtCore import QRunnable, QThreadPool

from kast.interface.qt5.service.InterfaceScheduler import InterfaceScheduler
from kast.interface.qt5.utils.dialogs import dialogCriticalError
from kast.utils.Loggable import Loggable

Callback = Callable[[], None]


class Worker(Loggable, QRunnable):

    def __init__(self, interfaceScheduler: InterfaceScheduler, callback: Callback) -> None:
        super().__init__()
        self._interfaceScheduler = interfaceScheduler
        self._callback = callback

    def run(self) -> None:
        try:
            self._callback()
        except Exception as ex:
            self.log.critical(f"Unhandled error in background task! Exception: {ex}", exc_info=True)

            def notifyCritical() -> None:
                try:
                    message = "Application encountered a CRITICAL error!" \
                          "\n(Shutting down due to high severity.)"
                    dialogCriticalError(message=message)
                finally:
                    sys.exit(1)

            self._interfaceScheduler.schedule(notifyCritical)


class BackgroundRunner:

    def __init__(self, interfaceScheduler: InterfaceScheduler) -> None:
        self._interfaceScheduler = interfaceScheduler
        self._threadPool = QThreadPool()

    def execute(self, callback: Callback) -> None:
        self._threadPool.start(Worker(interfaceScheduler=self._interfaceScheduler, callback=callback))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from typing import List

from kast.media.processing.SubtitlesSource import ISubtitlesSource


class SubtitlesSourceList:

    def __init__(self) -> None:
        self._sources: List[ISubtitlesSource] = []

    def __len__(self) -> int:
        return len(self._sources)

    def __getitem__(self, index: int) -> ISubtitlesSource:
        return self._sources[index]

    def clear(self) -> None:
        self._sources = []

    def append(self, source: ISubtitlesSource) -> None:
        self._sources.append(source)

    def pop(self, index: int) -> ISubtitlesSource:
        return self._sources.pop(index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

__package_name__ = __name__
__app_name__ = 'Kast'
__app_description__ = "Cast videos (with subtitles) straight from your PC."
__app_url__ = "https://bitbucket.org/massultidev/kast/"

__version_info__ = (1, 1, 2)
__version__ = '.'.join([str(i) for i in __version_info__])

__author__ = 'P.J. Grochowski'
__email__ = 'pawel.grochowski.dev@gmail.com'



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pkg/kast/interface/qt5/view/MainWindowView.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindowView(object):
    def setupUi(self, MainWindowView):
        MainWindowView.setObjectName("MainWindowView")
        MainWindowView.resize(800, 600)
        MainWindowView.setMinimumSize(QtCore.QSize(800, 600))
        self._widget = QtWidgets.QWidget(MainWindowView)
        self._widget.setObjectName("_widget")
        self._layout = QtWidgets.QVBoxLayout(self._widget)
        self._layout.setObjectName("_layout")
        self._centralLayout = QtWidgets.QVBoxLayout()
        self._centralLayout.setSpacing(0)
        self._centralLayout.setObjectName("_centralLayout")
        self.layoutForPreview = QtWidgets.QVBoxLayout()
        self.layoutForPreview.setSpacing(0)
        self.layoutForPreview.setObjectName("layoutForPreview")
        self._centralLayout.addLayout(self.layoutForPreview)
        self.layoutForControls = QtWidgets.QVBoxLayout()
        self.layoutForControls.setSpacing(0)
        self.layoutForControls.setObjectName("layoutForControls")
        self._centralLayout.addLayout(self.layoutForControls)
        self._centralLayout.setStretch(0, 1)
        self._layout.addLayout(self._centralLayout)
        MainWindowView.setCentralWidget(self._widget)
        self.menuBar = QtWidgets.QMenuBar(MainWindowView)
        self.menuBar.setGeometry(QtCore.QRect(0, 0, 800, 30))
        self.menuBar.setObjectName("menuBar")
        self.menuApplication = QtWidgets.QMenu(self.menuBar)
        self.menuApplication.setObjectName("menuApplication")
        MainWindowView.setMenuBar(self.menuBar)
        self.actionAbout = QtWidgets.QAction(MainWindowView)
        self.actionAbout.setObjectName("actionAbout")
        self.actionExit = QtWidgets.QAction(MainWindowView)
        self.actionExit.setObjectName("actionExit")
        self.menuApplication.addAction(self.actionAbout)
        self.menuApplication.addAction(self.actionExit)
        self.menuBar.addAction(self.menuApplication.menuAction())

        self.retranslateUi(MainWindowView)
        QtCore.QMetaObject.connectSlotsByName(MainWindowView)

    def retranslateUi(self, MainWindowView):
        _translate = QtCore.QCoreApplication.translate
        MainWindowView.setWindowTitle(_translate("MainWindowView", "MainWindowView"))
        self.menuApplication.setTitle(_translate("MainWindowView", "&Application"))
        self.actionAbout.setText(_translate("MainWindowView", "&About"))
        self.actionExit.setText(_translate("MainWindowView", "E&xit"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pkg/kast/interface/qt5/view/StatusBarView.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_StatusBarView(object):
    def setupUi(self, StatusBarView):
        StatusBarView.setObjectName("StatusBarView")
        StatusBarView.resize(660, 60)
        self.verticalLayout = QtWidgets.QVBoxLayout(StatusBarView)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self._separator = QtWidgets.QFrame(StatusBarView)
        self._separator.setFrameShape(QtWidgets.QFrame.HLine)
        self._separator.setFrameShadow(QtWidgets.QFrame.Sunken)
        self._separator.setObjectName("_separator")
        self.verticalLayout.addWidget(self._separator)
        self._layout = QtWidgets.QHBoxLayout()
        self._layout.setContentsMargins(-1, 3, -1, -1)
        self._layout.setSpacing(0)
        self._layout.setObjectName("_layout")
        self._layoutStatus = QtWidgets.QHBoxLayout()
        self._layoutStatus.setContentsMargins(-1, 6, -1, 6)
        self._layoutStatus.setSpacing(3)
        self._layoutStatus.setObjectName("_layoutStatus")
        self._labelStatus = QtWidgets.QLabel(StatusBarView)
        self._labelStatus.setObjectName("_labelStatus")
        self._layoutStatus.addWidget(self._labelStatus)
        self.labelStatusMessage = QtWidgets.QLabel(StatusBarView)
        self.labelStatusMessage.setObjectName("labelStatusMessage")
        self._layoutStatus.addWidget(self.labelStatusMessage)
        self._layoutStatus.setStretch(1, 1)
        self._layout.addLayout(self._layoutStatus)
        self.widgetProgress = QtWidgets.QWidget(StatusBarView)
        self.widgetProgress.setObjectName("widgetProgress")
        self._layoutProgress = QtWidgets.QHBoxLayout(self.widgetProgress)
        self._layoutProgress.setContentsMargins(0, 0, 0, 0)
        self._layoutProgress.setObjectName("_layoutProgress")
        self.progressBar = QtWidgets.QProgressBar(self.widgetProgress)
        self.progressBar.setProperty("value", 0)
        self.progressBar.setInvertedAppearance(False)
        self.progressBar.setObjectName("progressBar")
        self._layoutProgress.addWidget(self.progressBar)
        self.buttonCancel = QtWidgets.QPushButton(self.widgetProgress)
        self.buttonCancel.setText("")
        self.buttonCancel.setObjectName("buttonCancel")
        self._layoutProgress.addWidget(self.buttonCancel)
        self._layoutProgress.setStretch(0, 1)
        self._layout.addWidget(self.widgetProgress)
        self.verticalLayout.addLayout(self._layout)

        self.retranslateUi(StatusBarView)
        QtCore.QMetaObject.connectSlotsByName(StatusBarView)

    def retranslateUi(self, StatusBarView):
        _translate = QtCore.QCoreApplication.translate
        StatusBarView.setWindowTitle(_translate("StatusBarView", "StatusBarView"))
        self._labelStatus.setText(_translate("StatusBarView", "Status:"))
        self.labelStatusMessage.setText(_translate("StatusBarView", "Idle"))
        self.buttonCancel.setToolTip(_translate("StatusBarView", "Cancel"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from pathlib import Path
from types import ModuleType


class ResourceProvider:

    def __init__(self, package: ModuleType, assetsDirName: str = 'assets') -> None:
        self.assetsDir = Path(package.__file__).parent.joinpath(assetsDirName)

    def getResourcePath(self, assetRelativePath: Path) -> Path:
        return self.assetsDir.joinpath(assetRelativePath)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from PyQt5.QtDBus import QDBusConnection, QDBusMessage
from PyQt5.QtGui import QWindow

from kast.interface.qt5.UiServices import UiServices
from kast.interface.qt5.service.UiEvent import UiEvent, UiState
from kast.interface.qt5.viewModel.taskbar.TaskbarViewModelBase import TaskbarViewModelBase
from kast.media.casting.CastState import CastMediaState


class UnityTaskbarViewModel(TaskbarViewModelBase):

    def __init__(self, parent: QWindow, uiServices: UiServices) -> None:
        super().__init__(parent, uiServices)
        self.uiServices.uiEventObserver.register(self, self._onUiEvent)

        self._displayProgress(False)

    @property
    def _mediaState(self) -> CastMediaState:
        return self.uiServices.uiStateService.castState.mediaState

    def _onUiEvent(self, uiEvent: UiEvent) -> None:
        self._updateProgress(uiEvent)

    def _updateProgress(self, uiEvent: UiEvent) -> None:
        if not uiEvent.progress.complete:
            percentage = uiEvent.progress.percentage
            percentage = percentage/100 if percentage is not None else 1.0
            self._displayProgress(True, percentage)
            return

        duration = int(self._mediaState.duration)
        if (uiEvent.state == UiState.Streaming) and (duration > 0):
            self._displayProgress(True, int(self._mediaState.currentTime)/duration)
            return

        self._displayProgress(False)

    def _displayProgress(self, display: bool, percentage: float = 0.0) -> None:
        self._sendMessage({
            'progress': percentage,
            'progress-visible': display
        })

    def _sendMessage(self, parameters: dict) -> None:
        message = QDBusMessage.createSignal('/', 'com.canonical.Unity.LauncherEntry', 'Update')
        message << f'application://{self.services.settings.desktopFileName}'
        message << parameters
        QDBusConnection.sessionBus().send(message)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'pkg/kast/interface/qt5/view/VideoSettingsView.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_VideoSettingsView(object):
    def setupUi(self, VideoSettingsView):
        VideoSettingsView.setObjectName("VideoSettingsView")
        VideoSettingsView.resize(604, 173)
        VideoSettingsView.setWindowTitle("VideoSettingsView")
        self._layout = QtWidgets.QVBoxLayout(VideoSettingsView)
        self._layout.setObjectName("_layout")
        self._layoutMain = QtWidgets.QWidget(VideoSettingsView)
        self._layoutMain.setObjectName("_layoutMain")
        self.verticalLayout = QtWidgets.QVBoxLayout(self._layoutMain)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self._layoutDeviceAndVideo = QtWidgets.QGridLayout()
        self._layoutDeviceAndVideo.setContentsMargins(0, 0, 0, 10)
        self._layoutDeviceAndVideo.setHorizontalSpacing(10)
        self._layoutDeviceAndVideo.setVerticalSpacing(6)
        self._layoutDeviceAndVideo.setObjectName("_layoutDeviceAndVideo")
        self.lineEditVideo = QtWidgets.QLineEdit(self._layoutMain)
        self.lineEditVideo.setEnabled(True)
        self.lineEditVideo.setReadOnly(True)
        self.lineEditVideo.setObjectName("lineEditVideo")
        self._layoutDeviceAndVideo.addWidget(self.lineEditVideo, 1, 1, 1, 1)
        self._labelAudio = QtWidgets.QLabel(self._layoutMain)
        self._labelAudio.setObjectName("_labelAudio")
        self._layoutDeviceAndVideo.addWidget(self._labelAudio, 2, 0, 1, 1)
        self._labelDevice = QtWidgets.QLabel(self._layoutMain)
        self._labelDevice.setObjectName("_labelDevice")
        self._layoutDeviceAndVideo.addWidget(self._labelDevice, 0, 0, 1, 1)
        self.buttonSubtitlesAdd = QtWidgets.QPushButton(self._layoutMain)
        self.buttonSubtitlesAdd.setText("")
        self.buttonSubtitlesAdd.setObjectName("buttonSubtitlesAdd")
        self._layoutDeviceAndVideo.addWidget(self.buttonSubtitlesAdd, 2, 3, 1, 1)
        self._layoutAudioAndSubtitles = QtWidgets.QGridLayout()
        self._layoutAudioAndSubtitles.setContentsMargins(0, 0, 0, 0)
        self._layoutAudioAndSubtitles.setHorizontalSpacing(10)
        self._layoutAudioAndSubtitles.setVerticalSpacing(0)
        self._layoutAudioAndSubtitles.setObjectName("_layoutAudioAndSubtitles")
        self._labelSubtitles = QtWidgets.QLabel(self._layoutMain)
        self._labelSubtitles.setObjectName("_labelSubtitles")
        self._layoutAudioAndSubtitles.addWidget(self._labelSubtitles, 0, 2, 1, 1)
        self.comboBoxAudio = QtWidgets.QComboBox(self._layoutMain)
        self.comboBoxAudio.setObjectName("comboBoxAudio")
        self._layoutAudioAndSubtitles.addWidget(self.comboBoxAudio, 0, 1, 1, 1)
        self.comboBoxSubtitles = QtWidgets.QComboBox(self._layoutMain)
        self.comboBoxSubtitles.setObjectName("comboBoxSubtitles")
        self._layoutAudioAndSubtitles.addWidget(self.comboBoxSubtitles, 0, 3, 1, 1)
        self._layoutAudioAndSubtitles.setColumnStretch(1, 1)
        self._layoutAudioAndSubtitles.setColumnStretch(3, 1)
        self._layoutDeviceAndVideo.addLayout(self._layoutAudioAndSubtitles, 2, 1, 1, 1)
        self.comboBoxDevice = QtWidgets.QComboBox(self._layoutMain)
        self.comboBoxDevice.setObjectName("comboBoxDevice")
        self._layoutDeviceAndVideo.addWidget(self.comboBoxDevice, 0, 1, 1, 1)
        self.buttonDeviceRefresh = QtWidgets.QPushButton(self._layoutMain)
        self.buttonDeviceRefresh.setText("")
        self.buttonDeviceRefresh.setObjectName("buttonDeviceRefresh")
        self._layoutDeviceAndVideo.addWidget(self.buttonDeviceRefresh, 0, 3, 1, 1)
        self._labelVideo = QtWidgets.QLabel(self._layoutMain)
        self._labelVideo.setObjectName("_labelVideo")
        self._layoutDeviceAndVideo.addWidget(self._labelVideo, 1, 0, 1, 1)
        self.buttonVideoOpen = QtWidgets.QPushButton(self._layoutMain)
        self.buttonVideoOpen.setText("")
        self.buttonVideoOpen.setObjectName("buttonVideoOpen")
        self._layoutDeviceAndVideo.addWidget(self.buttonVideoOpen, 1, 3, 1, 1)
        self.verticalLayout.addLayout(self._layoutDeviceAndVideo)
        self._layoutStream = QtWidgets.QHBoxLayout()
        self._layoutStream.setObjectName("_layoutStream")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self._layoutStream.addItem(spacerItem)
        self.buttonStream = QtWidgets.QPushButton(self._layoutMain)
        self.buttonStream.setObjectName("buttonStream")
        self._layoutStream.addWidget(self.buttonStream)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self._layoutStream.addItem(spacerItem1)
        self.verticalLayout.addLayout(self._layoutStream)
        self._layout.addWidget(self._layoutMain)

        self.retranslateUi(VideoSettingsView)
        QtCore.QMetaObject.connectSlotsByName(VideoSettingsView)

    def retranslateUi(self, VideoSettingsView):
        _translate = QtCore.QCoreApplication.translate
        self.lineEditVideo.setToolTip(_translate("VideoSettingsView", "Video file path"))
        self._labelAudio.setText(_translate("VideoSettingsView", "Audio:"))
        self._labelDevice.setText(_translate("VideoSettingsView", "Device:"))
        self.buttonSubtitlesAdd.setToolTip(_translate("VideoSettingsView", "Open subtitles"))
        self._labelSubtitles.setText(_translate("VideoSettingsView", "Subtitles:"))
        self.comboBoxAudio.setToolTip(_translate("VideoSettingsView", "Audio tracks"))
        self.comboBoxSubtitles.setToolTip(_translate("VideoSettingsView", "Subtitles"))
        self.comboBoxDevice.setToolTip(_translate("VideoSettingsView", "Cast devices"))
        self.buttonDeviceRefresh.setToolTip(_translate("VideoSettingsView", "Search for cast devices"))
        self._labelVideo.setText(_translate("VideoSettingsView", "Video:"))
        self.buttonVideoOpen.setToolTip(_translate("VideoSettingsView", "Open video"))
        self.buttonStream.setToolTip(_translate("VideoSettingsView", "Start streaming"))
        self.buttonStream.setText(_translate("VideoSettingsView", "<< Stream >>"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from PyQt5.QtCore import QThread
from PyQt5.QtWidgets import QApplication

from kast.interface.qt5.utils.Schedulable import ISchedulable


class SchedulableException(Exception):
    pass


def foregroundTask(func):
    """Decorator used to make sure method runs in interface thread. Decorated methods must return None!"""
    def wrapper(self: ISchedulable, *args, **kwargs):
        _verifySchedulable(self)

        def callback() -> None:
            func(self, *args, **kwargs)

        if _isMainThread():
            callback()
            return

        self.schedulers.interfaceScheduler.schedule(callback)

    return wrapper


def backgroundTaskEx(forceSpawnThread: bool = False):
    """Decorator used to make sure method runs in background thread. Decorated methods must return None!"""
    def decorator(func):
        def wrapper(self: ISchedulable, *args, **kwargs):
            _verifySchedulable(self)

            def callback() -> None:
                func(self, *args, **kwargs)

            if not forceSpawnThread and not _isMainThread():
                callback()
                return

            self.schedulers.backgroundRunner.execute(lambda: self.backgroundTaskMiddleware(callback))

        return wrapper

    return decorator


backgroundTask = backgroundTaskEx()


def _isMainThread() -> bool:
    return QThread.currentThread() == QApplication.instance().thread()


def _verifySchedulable(obj):
    if not isinstance(obj, ISchedulable):
        raise SchedulableException("Non schedulable service cannot use schedulable decorators!")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import time

from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QStyle, QWidget

from kast.interface.qt5.UiServices import UiServices
from kast.interface.qt5.service.UiEvent import UiEvent, UiState
from kast.interface.qt5.view.MediaControlView import Ui_MediaControlView
from kast.interface.qt5.viewModel.VideoPreviewViewModel import VideoPreviewViewModel
from kast.interface.qt5.viewModel.ViewModelBase import ViewBase, ViewModelBase
from kast.media.casting.CastState import CastMediaState, CastPlayerState, CastState, Seconds
from kast.utils.chrono import Milliseconds


class View(ViewBase, QWidget, Ui_MediaControlView):
    pass


class MediaControlViewModel(ViewModelBase):

    def __init__(
        self,
        parent: QWidget,
        uiServices: UiServices,
        videoPreviewViewModel: VideoPreviewViewModel
    ) -> None:
        self._view = View(parent=parent)
        super().__init__(uiServices=uiServices, view=self._view)
        self.uiServices.uiEventObserver.register(self, self._onUiEvent)

        self._videoPreviewViewModel = videoPreviewViewModel
        self._videoPreviewViewModel.setMuteCallback(self._onLocalMute)
        self._videoPreviewViewModel.setVolumeCallback(self._onLocalVolume)

        self._view.buttonPlayPause.setIcon(self._view.style().standardIcon(QStyle.SP_MediaPlay))
        self._view.buttonStop.setIcon(self._view.style().standardIcon(QStyle.SP_MediaStop))
        self._view.buttonSeekFront.setIcon(self._view.style().standardIcon(QStyle.SP_MediaSeekForward))
        self._view.buttonSeekBack.setIcon(self._view.style().standardIcon(QStyle.SP_MediaSeekBackward))
        self._view.buttonMuteRemote.setIcon(self._view.style().standardIcon(QStyle.SP_MediaVolume))
        self._view.buttonMuteLocal.setIcon(self._view.style().standardIcon(QStyle.SP_MediaVolume))

        self._videoPositionUpdater = QTimer()
        self._videoPositionUpdater.setInterval(int(Milliseconds(Seconds(1))))
        self._videoPositionUpdater.timeout.connect(self._signalVideoPositionUpdate)
        self._videoPositionUpdater.start()

        self._view.buttonDisconnect.clicked.connect(self._signalClickedDisconnect)
        self._view.buttonPlayPause.clicked.connect(self._signalClickedPlayPause)
        self._view.buttonStop.clicked.connect(self._signalClickedStop)
        self._view.buttonSeekFront.clicked.connect(self._signalClickedSeekForward)
        self._view.buttonSeekBack.clicked.connect(self._signalClickedSeekBackward)
        self._view.buttonMuteRemote.clicked.connect(self._signalClickedMuteRemote)
        self._view.buttonMuteLocal.clicked.connect(self._signalClickedMuteLocal)

        self._view.sliderSeek.sliderReleased.connect(self._signalSeekPosition)
        self._view.sliderVolumeRemote.sliderReleased.connect(self._signalSetVolumeRemote)
        self._view.sliderVolumeLocal.sliderReleased.connect(self._signalSetVolumeLocal)

        self._view.setVisible(False)

    @property
    def _castState(self) -> CastState:
        return self.uiServices.uiStateService.castState

    @property
    def _mediaState(self) -> CastMediaState:
        return self._castState.mediaState

    def _onUiEvent(self, uiEvent: UiEvent) -> None:
        isStreaming = uiEvent.state == UiState.Streaming
        self._view.setVisible(isStreaming)
        self._videoPreviewViewModel.enable(state=isStreaming)
        self._videoPreviewViewModel.setPlayerState(state=self._mediaState.playerState)

        if uiEvent.state == UiState.Closing:
            self._videoPreviewViewModel.cleanup()

        iconPlayPause = QStyle.SP_MediaPause if self._mediaState.isPlaying() else QStyle.SP_MediaPlay
        self._view.buttonPlayPause.setIcon(self._view.style().standardIcon(iconPlayPause))

        self._updateVideoPosition(position=self._mediaState.currentTime, duration=self._mediaState.duration)

        volumeLevel = int(round(self._mediaState.volumeLevel * 100))
        self._view.sliderVolumeRemote.setSliderPosition(volumeLevel)

        iconMuted = QStyle.SP_MediaVolumeMuted if self._mediaState.volumeMuted or volumeLevel == 0 else QStyle.SP_MediaVolume
        self._view.buttonMuteRemote.setIcon(self._view.style().standardIcon(iconMuted))

    def _onLocalMute(self, muted: bool) -> None:
        iconMuted = QStyle.SP_MediaVolumeMuted if muted else QStyle.SP_MediaVolume
        self._view.buttonMuteLocal.setIcon(self._view.style().standardIcon(iconMuted))

    def _onLocalVolume(self, volume: int) -> None:
        self._view.sliderVolumeLocal.setValue(volume)

    def _updateVideoPosition(self, position: Seconds, duration: Seconds) -> None:
        def formatTime(seconds: Seconds) -> str:
            return time.strftime('%H:%M:%S', time.gmtime(seconds.value))

        self._view.labelTime.setText(f"{formatTime(position)} / {formatTime(duration)}")

        self._view.sliderSeek.setRange(0, duration.value)
        self._view.sliderSeek.setSliderPosition(position.value)

        self._videoPreviewViewModel.updateVideoPosition(position=position)

    def _signalVideoPositionUpdate(self) -> None:
        if self._mediaState.playerState == CastPlayerState.Playing:
            self._updateVideoPosition(
                position=self._mediaState.currentTime,
                duration=self._mediaState.duration
            )

    def _signalClickedDisconnect(self) -> None:
        self.services.castController.quit()
        self.services.castController.disconnect()

    def _signalClickedPlayPause(self) -> None:
        self.uiServices.mediaControlService.playOrPause()

    def _signalClickedStop(self) -> None:
        self.uiServices.mediaControlService.stop()

    def _signalClickedSeekForward(self) -> None:
        self.uiServices.mediaControlService.seekForward()

    def _signalClickedSeekBackward(self) -> None:
        self.uiServices.mediaControlService.seekBackward()

    def _signalSeekPosition(self) -> None:
        self.uiServices.mediaControlService.seek(Seconds(self._view.sliderSeek.value()))

    def _signalClickedMuteRemote(self) -> None:
        self.uiServices.mediaControlService.setMute(not self._mediaState.volumeMuted)

    def _signalSetVolumeRemote(self) -> None:
        self.uiServices.mediaControlService.setVolume(self._view.sliderVolumeRemote.value()/100)

    def _signalClickedMuteLocal(self) -> None:
        self._videoPreviewViewModel.muted = not self._videoPreviewViewModel.muted

    def _signalSetVolumeLocal(self) -> None:
        self._videoPreviewViewModel.volume = self._view.sliderVolumeLocal.value()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import threading
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional

from kast.Services import Services
from kast.interface.qt5.service.BackgroundRunner import BackgroundRunner
from kast.interface.qt5.service.InterfaceScheduler import InterfaceScheduler
from kast.interface.qt5.service.UiEvent import Progress, UiEvent, UiState
from kast.interface.qt5.service.UiStateService import UiStateService
from kast.interface.qt5.utils.Schedulable import Callback, Schedulable
from kast.interface.qt5.utils.dialogs import dialogError, dialogQuestionOkCancel
from kast.interface.qt5.utils.schedule import backgroundTask, foregroundTask
from kast.media.casting.CastException import CastException
from kast.media.casting.CastState import CastMediaState, CastState, DeviceName
from kast.media.processing.MetaData import MetaData
from kast.media.processing.SubtitleUtils import SubtitleException
from kast.media.processing.SubtitlesSource import SubtitlesFromFile, SubtitlesFromStream
from kast.media.processing.SubtitlesSourceList import SubtitlesSourceList
from kast.media.processing.Transcoder import Codecs, Streams, TranscodeParams, Transcoder
from kast.media.processing.common import StreamId, containerExtension
from kast.utils.Loggable import Loggable
from kast.utils.chrono import Seconds


@dataclass
class StreamParams:
    deviceName: DeviceName
    audioStreamId: StreamId
    subtitlesId: Optional[int] = None


class MediaControlService(Loggable, Schedulable):

    VideoOpenedCallback = Callable[[MetaData], None]
    TaskCallback = Callable[[], None]

    def __init__(
        self,
        services: Services,
        interfaceScheduler: InterfaceScheduler,
        backgroundRunner: BackgroundRunner,
        uiStateService: UiStateService
    ) -> None:
        super().__init__(
            interfaceScheduler=interfaceScheduler,
            backgroundRunner=backgroundRunner
        )

        self._services = services
        self._interfaceScheduler = interfaceScheduler
        self._backgroundRunner = backgroundRunner
        self._uiStateService = uiStateService

        self._videoFilePath: Optional[Path] = None
        self._metaData = MetaData()
        self._subtitlesSourceList = SubtitlesSourceList()
        self._lastStreamParams: Optional[StreamParams] = None
        self._lastTranscodeParams: Optional[TranscodeParams] = None

    @property
    def _castState(self) -> CastState:
        return self._uiStateService.castState

    @property
    def _mediaState(self) -> CastMediaState:
        return self._castState.mediaState

    def playOrPause(self) -> None:
        if self._mediaState.isPlaying():
            self._services.castController.pause()
        else:
            self._services.castController.play()

    def stop(self) -> None:
        self._services.castController.stop()  # TODO: Make it possible to replay movie!

    def seekForward(self) -> None:
        self._services.castController.seekForward()

    def seekBackward(self) -> None:
        self._services.castController.seekBackward()

    def seek(self, position: Seconds) -> None:
        self._services.castController.seek(position)

    def setMute(self, mute: bool = True) -> None:
        self._services.castController.setMute(mute)

    def setVolume(self, value: float) -> None:
        self._services.castController.setVolume(value)

    def openVideo(self, filePath: Path, callback: VideoOpenedCallback) -> None:
        self._videoFilePath = filePath

        self._saveLastMediaDir(filePath)

        self._uiStateService.dispatch(UiEvent(state=UiState.VideoProbe, progress=Progress(complete=False)))

        self._extractMetaData(filePath, callback)

    @backgroundTask
    def _extractMetaData(self, filePath: Path, callback: VideoOpenedCallback) -> None:
        metaData = self._services.mediaProcessingService.extractMetaData(inputFile=filePath)

        self._notifyMetaDataExtracted(metaData, filePath, callback)

    @foregroundTask
    def _notifyMetaDataExtracted(self, metaData: MetaData, filePath: Path, callback: VideoOpenedCallback) -> None:
        self._metaData = metaData

        self._subtitlesSourceList.clear()
        for streamId in range(len(metaData.subtitleStreamLangs)):
            self._subtitlesSourceList.append(SubtitlesFromStream(
                mediaProcessingService=self._services.mediaProcessingService,
                mediaFile=filePath,
                streamId=streamId
            ))

        callback(metaData)

        self._uiStateService.dispatch(UiEvent(state=UiState.Idle))

    def addSubtitles(self, filePath: Path) -> None:
        self._saveLastMediaDir(filePath)

        self._subtitlesSourceList.append(SubtitlesFromFile(
            mediaProcessingService=self._services.mediaProcessingService,
            subtitlesFile=filePath
        ))

    def startStream(
        self,
        deviceName: DeviceName,
        audioStreamId: StreamId,
        subtitlesId: Optional[int] = None
    ) -> None:
        if not self._videoFilePath:
            dialogError("No video file has been selected!")
            return

        self._lastStreamParams = StreamParams(deviceName, audioStreamId, subtitlesId)

        self._uiStateService.dispatch(UiEvent(state=UiState.Preprocessing, progress=Progress(complete=False)))

        self._initTranscoding()

    @backgroundTask
    def _initTranscoding(self) -> None:
        cancelEvent = threading.Event()

        def progressCallback(percent: int, complete: bool) -> None:
            self.log.info(f"Transcoding progress: {percent}% ({'Complete' if complete else 'Running'})")
            self._uiStateService.dispatch(UiEvent(state=UiState.AvProcessing, progress=Progress(
                complete=complete, percentage=percent, cancelEvent=cancelEvent
            )))

        castController = self._services.castController

        inputContainer = containerExtension(self._videoFilePath)
        outputContainer = inputContainer if inputContainer in castController.supportedContainerFormats \
            else castController.preferredContainerFormat

        outputCodecNames = Codecs(
            video=castController.preferredVideoCodec,
            audio=castController.preferredAudioCodec
        )

        transcoder = self._services.mediaProcessingService.createTranscoder(
            inputFile=self._videoFilePath,
            inputStreamIds=Streams(video=0, audio=self._lastStreamParams.audioStreamId),
            outputCodecNames=outputCodecNames,
            containerFormat=outputContainer,
            progressCallback=progressCallback,
            cancelEvent=cancelEvent
        )
        self.log.info(
            "Input file: "
            f"container='{inputContainer}', "
            f"codecs={transcoder.inputCodecNames}, "
            f"resolution={transcoder.inputResolution}"
        )
        self.log.info(
            "Output file: "
            f"container='{outputContainer}', "
            f"codecs={transcoder.outputCodecNames}, "
            f"resolution={transcoder.outputResolution}"
        )
        if transcoder.requireProcessing and self._lastTranscodeParams != transcoder.params:
            self._confirmTranscoding(transcoder)
            return

        self._startStreaming(transcoder.outputFile)

    @foregroundTask
    def _confirmTranscoding(self, transcoder: Transcoder) -> None:
        message = "Selected media is not supported in it's current form. " \
            "Either because of a codec mismatch, too high resolution or not supported media container type. " \
            "Your file can be remuxed and transcoded to address those issues.\n" \
            "\n" \
            "Remuxing is a quick process. While duration of transcoding varies. " \
            "It is usually fast for audio codecs and can take some time for video codecs. " \
            "Actual times will depend on your machine processing power.\n" \
            "\n" \
            "Provided Container -> Supported Container:\n" \
            f"- {containerExtension(transcoder.inputFile)} -> {containerExtension(transcoder.outputFile)}\n" \
            "\n" \
            "Provided Codecs -> Supported Codecs:\n" \
            f"- Video: '{transcoder.inputCodecNames.video}' -> '{transcoder.outputCodecNames.video}'\n" \
            f"- Audio: '{transcoder.inputCodecNames.audio}' -> '{transcoder.outputCodecNames.audio}'\n" \
            "\n" \
            f"Provided resolution / Max supported resolution:\n" \
            f"- '{transcoder.inputResolution}' / '{transcoder.maxResolution}'\n" \
            "\n" \
            "Proceed with media processing?\n" \
            "(Your original file will not be modified.)\n"

        if not dialogQuestionOkCancel(title="Media Processing", message=message):
            self._cancelAction()
            return

        self._startTranscoding(transcoder)

    @backgroundTask
    def _startTranscoding(self, transcoder: Transcoder) -> None:
        if not transcoder.run():
            self._cancelAction()
            return

        self._lastTranscodeParams = transcoder.params

        self._confirmStreaming(transcoder.outputFile)

    @foregroundTask
    def _confirmStreaming(self, videoFile: Path) -> None:
        message = "Media processing finished!\n\nProceed with streaming?\n"

        if not dialogQuestionOkCancel(title="Streaming", message=message):
            self._cancelAction()
            return

        self._startStreaming(videoFile)

    @backgroundTask
    def _startStreaming(self, videoFile: Path) -> None:
        mediaContent = self._services.mediaServer.mediaContent
        mediaContent.movieFile = videoFile
        mediaContent.subtitlesFile = self._getSubtitlesFilePath()

        self._uiStateService.dispatch(UiEvent(state=UiState.Connecting, progress=Progress(complete=False)))

        self._services.castController.connect(name=self._lastStreamParams.deviceName)
        self._services.castController.stream(
            movieUrl=self._services.mediaServer.movieUrl,
            subtitlesUrl=self._services.mediaServer.subtitleUrl,
            thumbnailUrl=self._services.mediaServer.thumbnailUrl,
            title=self._metaData.title
        )

        self._uiStateService.dispatch(UiEvent(state=UiState.Streaming))

    def _getSubtitlesFilePath(self) -> Optional[Path]:
        subtitlesId = self._lastStreamParams.subtitlesId
        return self._subtitlesSourceList[subtitlesId].toVtt() if subtitlesId is not None else None

    def _cancelAction(self) -> None:
        self._uiStateService.dispatch(UiEvent(state=UiState.Idle))

    def _saveLastMediaDir(self, filePath) -> None:
        self._services.settings.browseMediaDir = filePath.parent

    def _reportError(self, message) -> None:
        def interfaceCallback() -> None:
            dialogError(message=message)
            self._uiStateService.dispatch(UiEvent(state=UiState.Idle), synchronous=True)

        self._interfaceScheduler.schedule(interfaceCallback)

    def backgroundTaskMiddleware(self, callback: Callback) -> None:
        try:
            callback()
        except (SubtitleException, CastException) as ex:
            message = str(ex)  # Workaround for a 'free variable referenced before assignment' problem.
            self.log.exception(ex)
            self._reportError(message)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from typing import List, Optional

from mpris_server.adapters import MprisAdapter
from mpris_server.base import DbusObj, Microseconds, PlayState, URI, VolumeDecimal
from mpris_server.compat import enforce_dbus_length, get_dbus_name
from mpris_server.metadata import Metadata

from kast.interface.overlay._mpris.common import castTimeToMprisTime, mprisTimeToCastTime
from kast.media.casting.CastController import CastController
from kast.media.casting.CastState import CastConnectionState, CastState
from kast.settings.Settings import Settings
from kast.settings.SettingsState import SettingsState


class MprisCastController(MprisAdapter):

    def __init__(
        self,
        settings: Settings,
        castController: CastController
    ) -> None:
        super().__init__(name=settings.appName)
        self._settings = settings
        self._castController = castController

        self._desktopFilePath = self._settings.persistentStoragePath.joinpath(f"{self._settings.package}.desktop")
        self._createDesktopFile()

    @property
    def _castState(self) -> CastState:
        return self._castController.castState

    def metadata(self) -> Metadata:

        @enforce_dbus_length
        def get_track_id(name: str) -> DbusObj:
            return f'/track/{get_dbus_name(name)}'

        title = self.get_stream_title()
        contentUrl = self._castState.mediaState.contentUrl
        contentUrl = contentUrl if contentUrl else None

        metadata = {
            'mpris:trackid': get_track_id(title),
            'mpris:length': castTimeToMprisTime(self._castState.mediaState.duration),
            'mpris:artUrl': self.get_art_url(),
            'xesam:url': contentUrl,
            'xesam:title': title,
            'xesam:artist': [self._castState.deviceName],
        }

        return metadata

    def get_art_url(self, track: Optional[int] = None) -> str:
        return self._castState.mediaState.imageUrl or \
            self._castState.mediaState.iconUrl or \
            str(self._settings.appIconPath)

    def get_desktop_entry(self) -> str:
        return str(self._desktopFilePath.with_suffix(suffix=''))

    def get_stream_title(self) -> str:
        title = self._castState.mediaState.title \
            if self._castState.mediaState.title is not None \
            else self._settings.appName

        if self._castState.connection == CastConnectionState.Lost:
            title = f'[Connection Lost] {title}'

        return title

    def get_uri_schemes(self) -> List[str]:
        return URI

    def get_mime_types(self) -> List[str]:
        return self._castController.supportedMimeTypes

    def can_quit(self) -> bool:
        return True

    def can_control(self) -> bool:
        return True

    def can_play(self) -> bool:
        return True

    def can_pause(self) -> bool:
        return self._castState.capabilities.canPause

    def can_seek(self) -> bool:
        return self._castState.capabilities.canSeek

    def can_go_next(self) -> bool:
        return False

    def can_go_previous(self) -> bool:
        return False

    def is_mute(self) -> bool:
        return self._castState.mediaState.volumeMuted

    def get_volume(self) -> VolumeDecimal:
        return self._castState.mediaState.volumeLevel

    def get_playstate(self) -> PlayState:
        if self._castState.mediaState.isPlaying():
            return PlayState.PLAYING
        if self._castState.mediaState.isPaused():
            return PlayState.PAUSED
        return PlayState.STOPPED

    def get_current_position(self) -> Microseconds:
        return castTimeToMprisTime(self._castState.mediaState.currentTime)

    def is_playlist(self) -> bool:
        return False

    def is_repeating(self) -> bool:
        return False

    def get_shuffle(self) -> bool:
        return False

    def quit(self) -> None:
        self._castController.quit()

    def set_mute(self, val: bool) -> None:
        self._castController.setMute(val)

    def set_volume(self, val: VolumeDecimal) -> None:
        self._castController.setVolume(val)

    def play(self) -> None:
        self._castController.play()

    def pause(self) -> None:
        self._castController.pause()

    def resume(self) -> None:
        self._castController.play()

    def stop(self) -> None:
        self._castController.stop()

    def seek(self, timePos: Microseconds) -> None:
        self._castController.seek(mprisTimeToCastTime(timePos))

    def next(self) -> None:
        pass

    def previous(self) -> None:
        pass

    def set_repeating(self, val: bool) -> None:
        pass

    def set_shuffle(self, val: bool) -> None:
        pass

    def set_loop_status(self, val: str) -> None:
        pass

    def onSettingsEvent(self, event: SettingsState) -> None:
        pass

    def _createDesktopFile(self) -> None:
        desktopFileContent = \
            "[Desktop Entry]\n" \
            "Encoding=UTF-8\n" \
            f"Name={self._settings.appName}\n" \
            f"Icon={self._settings.appIconPath}\n"

        with self._desktopFilePath.open(mode='w') as fOutput:
            fOutput.write(desktopFileContent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QDialog, QWidget

from kast.interface.qt5.UiServices import UiServices
from kast.interface.qt5.view.AboutViewModel import Ui_AboutViewModel
from kast.interface.qt5.viewModel.ViewModelBase import ViewBase, ViewModelBase


class View(ViewBase, QDialog, Ui_AboutViewModel):
    pass


class AboutViewModel(ViewModelBase):

    def __init__(self, parent: QWidget, uiServices: UiServices) -> None:
        self._view = View(parent=parent)
        super().__init__(uiServices=uiServices, view=self._view)
        self._view.hide()

        logo = QPixmap(str(self.services.settings.appIconPath)).scaled(128, 128)
        self._view.labelLogo.setPixmap(logo)

        self._view.labelAppName.setText(self.services.settings.appName)
        self._view.labelAppVersion.setText(self.services.settings.appVersion)
        self._view.labelAppDescription.setText("Cast movies (with subtitles) straight from your PC.")
        self._view.labelAuthorName.setText(self.services.settings.author)
        self._view.labelAuthorContact.setText(self.services.settings.email)

        self._view.layout().activate()
        self._view.setFixedSize(self._view.size().width(), self._view.size().height())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import kast
from kast.interface.overlay.MediaOverlay import MediaOverlay
from kast.media.casting.CastController import CastController
from kast.media.casting.CastEventObserver import CastEventObserver
from kast.media.processing.MediaProcessingService import MediaProcessingService
from kast.media.streaming.MediaServer import MediaServer
from kast.settings.Settings import Settings
from kast.settings.SettingsEventObserver import SettingsEventObserver
from kast.utils.ResourceProvider import ResourceProvider
from kast.utils.SingletonApplication import SingletonApplication


class Services:

    def __init__(self) -> None:
        self._appLock = SingletonApplication(appName=kast.__package_name__)
        self.resourceProvider = ResourceProvider(kast)
        self.settingsEventObserver = SettingsEventObserver()
        self.settings = Settings(resourceProvider=self.resourceProvider, settingsEventObserver=self.settingsEventObserver)

        self.castEventObserver = CastEventObserver()
        self.castController = CastController(settings=self.settings, castEventObserver=self.castEventObserver)
        self.mediaServer = MediaServer()
        self.mediaProcessingService = MediaProcessingService(self.settings.temporaryStoragePath)

        self.osMediaOverlay = MediaOverlay(settings=self.settings, castController=self.castController)
        self.settingsEventObserver.register(self.osMediaOverlay, self.osMediaOverlay.onSettingsEvent)
        self.castEventObserver.register(self.osMediaOverlay, self.osMediaOverlay.onCastEvent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import asyncio
from pathlib import Path
from typing import Tuple

from winrt import _winrt
from winrt.windows.foundation import Uri, TimeSpan
from winrt.windows.media import SystemMediaTransportControls, SystemMediaTransportControlsButtonPressedEventArgs, \
    SystemMediaTransportControlsButton, PlaybackPositionChangeRequestedEventArgs, MediaPlaybackType, \
    SystemMediaTransportControlsTimelineProperties, MediaPlaybackStatus
from winrt.windows.media.playback import MediaPlayer
from winrt.windows.storage import StorageFile
from winrt.windows.storage.streams import RandomAccessStreamReference

from kast.media.casting.CastController import CastController
from kast.media.casting.CastState import CastPlayerState
from kast.settings.Settings import Settings
from kast.utils.Loggable import Loggable
from kast.utils.chrono import Seconds, Milliseconds

# Workaround for COM thread configuration interference with PyQt:
_winrt.uninit_apartment()


class WinrtCastController(Loggable):

    def __init__(
        self,
        settings: Settings,
        castController: CastController
    ) -> None:
        self._settings = settings
        self._castController = castController

        self._mediaControls = mediaControls = self._createMediaControls()
        mediaControls.is_enabled = True

        mediaControls.add_button_pressed(lambda sender, args: self._onButtonPressed(sender=sender, args=args))
        mediaControls.add_playback_position_change_requested(lambda sender, args: self._onPositionChanged(sender=sender, args=args))

    def reload(self) -> None:
        self._updateControls()
        self._updateMetaData()
        self._updateTimeline()
        self._updatePlayerState()

    def _updateControls(self) -> None:
        enable = (not self._castController.castState.mediaState.isStopped())
        self._mediaControls.is_play_enabled = enable
        self._mediaControls.is_pause_enabled = enable
        self._mediaControls.is_stop_enabled = enable

    def _updateMetaData(self) -> None:
        title, subtitle = self._getTitleAndSubtitle()

        displayUpdater = self._mediaControls.display_updater
        displayUpdater.type = MediaPlaybackType.VIDEO
        displayUpdater.video_properties.title = title
        displayUpdater.video_properties.subtitle = subtitle
        displayUpdater.thumbnail = self._createThumbnailStream()

        displayUpdater.update()

    def _updateTimeline(self) -> None:
        mediaState = self._castController.castState.mediaState

        timelineProperties = SystemMediaTransportControlsTimelineProperties()
        timelineProperties.start_time = self._createTimeSpan()
        timelineProperties.min_seek_time = self._createTimeSpan()
        timelineProperties.position = self._createTimeSpan(mediaState.currentTime)
        timelineProperties.max_seek_time = self._createTimeSpan(mediaState.duration)
        timelineProperties.end_time = self._createTimeSpan(mediaState.duration)

        self._mediaControls.update_timeline_properties(timelineProperties)

    def _updatePlayerState(self) -> None:
        states = {
            CastPlayerState.Playing: MediaPlaybackStatus.PLAYING,
            CastPlayerState.Paused: MediaPlaybackStatus.PAUSED,
            CastPlayerState.Buffering: MediaPlaybackStatus.CHANGING,
            CastPlayerState.Idle: MediaPlaybackStatus.STOPPED,
            CastPlayerState.Unknown: MediaPlaybackStatus.CLOSED,
        }

        state = self._castController.castState.mediaState.playerState
        if state not in states:
            self.log.warning(f"Unhandled media player state change (state={state.name})!")
            return

        self._mediaControls.playback_status = states[state]

    def _onButtonPressed(
        self,
        sender: SystemMediaTransportControls,
        args: SystemMediaTransportControlsButtonPressedEventArgs
    ) -> None:
        button = SystemMediaTransportControlsButton(args.button)

        actions = {
            SystemMediaTransportControlsButton.PLAY: self._castController.play,
            SystemMediaTransportControlsButton.PAUSE: self._castController.pause,
            SystemMediaTransportControlsButton.STOP: self._castController.stop
        }

        if button not in actions:
            self.log.warning(f"Unhandled button press (button='{button.name}')!")
            return

        actions[button]()

    def _onPositionChanged(
        self,
        sender: SystemMediaTransportControls,
        args: PlaybackPositionChangeRequestedEventArgs
    ) -> None:
        position = Seconds(Milliseconds(args.requested_playback_position.duration))
        self._castController.seek(position)

    def _getTitleAndSubtitle(self) -> Tuple[str, str]:
        castState = self._castController.castState

        titles = []
        castState.mediaState.title and titles.append(castState.mediaState.title)
        titles.append(self._settings.appName + (f" ({castState.deviceName})" if castState.deviceName else ''))
        titles.append('')

        return titles[0], titles[1]

    def _createThumbnailStream(self) -> RandomAccessStreamReference:
        thumbnailUrl = self._castController.castState.mediaState.imageUrl or self._castController.castState.mediaState.iconUrl
        if thumbnailUrl:
            return RandomAccessStreamReference.create_from_uri(Uri(thumbnailUrl))

        return RandomAccessStreamReference.create_from_file(asyncio.run(self.openFile(self._settings.appIconPath)))

    @staticmethod
    def _createTimeSpan(seconds: Seconds = Seconds(0)) -> TimeSpan:
        return TimeSpan(Milliseconds(seconds).value)

    @staticmethod
    async def openFile(file: Path) -> StorageFile:
        return await StorageFile.get_file_from_path_async(str(file))

    @staticmethod
    def _createMediaControls() -> SystemMediaTransportControls:
        mediaPlayer = MediaPlayer()
        mediaPlayer.command_manager.is_enabled = False
        return mediaPlayer.system_media_transport_controls



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import copy

from kast.interface.qt5.service.InterfaceScheduler import InterfaceScheduler
from kast.interface.qt5.service.UiEvent import UiEvent, UiState
from kast.interface.qt5.service.UiEventObserver import UiEventObserver
from kast.media.casting.CastEventObserver import CastEventObserver
from kast.media.casting.CastState import CastState
from kast.utils.Loggable import Loggable


class UiStateService(Loggable):

    def __init__(
        self,
        interfaceScheduler: InterfaceScheduler,
        uiEventObserver: UiEventObserver,
        castEventObserver: CastEventObserver
    ) -> None:
        self._interfaceScheduler = interfaceScheduler
        self._uiEventObserver = uiEventObserver
        self._castEventObserver = castEventObserver

        self._uiEvent = UiEvent()
        self._castState = CastState()

        self._castEventObserver.register(listener=self, callback=self._onCastEvent)

    @property
    def uiEvent(self):
        return self._uiEvent

    @property
    def castState(self) -> CastState:
        return self._castState

    def dispatch(self, uiEvent: UiEvent, synchronous: bool = False) -> None:
        def interfaceCallback() -> None:
            self._uiEvent = uiEvent
            self._dispatch()

        if synchronous:
            interfaceCallback()
            return

        self._interfaceScheduler.schedule(interfaceCallback)

    def _onCastEvent(self, event: CastState) -> None:
        def interfaceCallback() -> None:
            self._castState = copy.deepcopy(event)
            self._updateUiState()
            self._dispatch()
        self._interfaceScheduler.schedule(interfaceCallback)

    def _updateUiState(self) -> None:
        if (self._uiEvent.state == UiState.Streaming) and (not self._castState.isConnectedOrRecoverable()):
            self._uiEvent = UiEvent(UiState.Idle)
            return

        if (self._uiEvent.state == UiState.Idle) and self._castState.isConnectedOrRecoverable():
            self._uiEvent = UiEvent(UiState.Streaming)
            return

    def _dispatch(self) -> None:
        self.log.info(
            f"UiState={self.uiEvent.state.name}, "
            f"DeviceState={self.castState.connection.value.lower().capitalize()}, "
            f"MediaState={self.castState.mediaState.playerState.value.lower().capitalize()}"
        )

        self._uiEventObserver.notify(self._uiEvent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from __future__ import annotations  # TODO: Remove when support dropped for: Python < 3.10

from dataclasses import dataclass
from functools import total_ordering


@total_ordering
@dataclass(frozen=True)
class Resolution:
    width: int
    height: int

    def __str__(self) -> str:
        return f"{self.width}x{self.height}"

    def __lt__(self, other: Resolution) -> bool:
        if not isinstance(other, Resolution):
            raise ValueError(f"Cannot compare {Resolution.__name__} with {other.__class__.__name__}!")

        return self.width < other.width or \
            self.height < other.height

    def shrinkToFit(self, boundingSize: Resolution) -> Resolution:
        def adjustDimension(dimToAdjust: int, dimOther: int, dimOtherBound: int) -> int:
            return int(dimToAdjust * (dimOtherBound / dimOther))

        def adjustWidth(resolution: Resolution) -> Resolution:
            if resolution.width < boundingSize.width:
                return resolution
            return Resolution(
                width=boundingSize.width,
                height=adjustDimension(
                    dimToAdjust=resolution.height,
                    dimOther=resolution.width,
                    dimOtherBound=boundingSize.width
                )
            )

        def adjustHeight(resolution: Resolution) -> Resolution:
            if resolution.height < boundingSize.height:
                return resolution
            return Resolution(
                width=adjustDimension(
                    dimToAdjust=resolution.width,
                    dimOther=resolution.height,
                    dimOtherBound=boundingSize.height
                ),
                height=boundingSize.height
            )

        return adjustHeight(adjustWidth(self))


FULL_HD = Resolution(1920, 1080)
ULTRA_HD = Resolution(3840, 2160)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from mpris_server.base import Microseconds

import kast.utils.chrono as chrono


def castTimeToMprisTime(value: chrono.Seconds) -> Microseconds:
    return chrono.Microseconds(value).value


def mprisTimeToCastTime(value: Microseconds) -> chrono.Seconds:
    return chrono.Seconds(chrono.Microseconds(value))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import os
import platform
import sys

from PyQt5 import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication

from kast.Services import Services
from kast.interface.qt5.UiServices import UiServices
from kast.interface.qt5.service.UiEvent import UiEvent, UiState, Progress
from kast.interface.qt5.viewModel.MainWindowViewModel import MainWindowViewModel
from kast.utils.Loggable import Loggable


class KastQtApp(Loggable):

    def __init__(self, services: Services) -> None:
        self._setupQt()
        self._app = QApplication(sys.argv)
        self._uiServices = UiServices(services)

        self._app.setWindowIcon(QIcon(str(services.settings.appIconPath)))
        self._setupPlatform()

        self._mainWindowViewModel = MainWindowViewModel(uiServices=self._uiServices)

        self._app.aboutToQuit.connect(self._onExit)

    def run(self) -> int:
        self.log.info(f"PyQt: {Qt.PYQT_VERSION_STR}, Qt: {Qt.QT_VERSION_STR}")
        return self._app.exec()

    def exit(self, returnCode: int = 0) -> None:
        self._app.closeAllWindows()
        self._app.exit(returnCode)

    def _onExit(self) -> None:
        self.log.info("Interface exit.")
        self._uiServices.uiEventObserver.notify(UiEvent(state=UiState.Closing, progress=Progress(complete=False)))

    def _setupPlatform(self) -> None:
        """Platform specific setup. Should be run after QApplication creation."""

        if platform.system() == 'Linux':
            self._app.setDesktopFileName(self._uiServices.services.settings.desktopFileName)

    @staticmethod
    def _setupQt() -> None:
        """Qt framework setup. Should be run before QApplication creation."""

        # On windows default backend for media support is 'directshow'.
        # Which does not provide support for proprietary codecs like 'h264'.
        # We could change to 'windowsmediafoundation' backend instead.
        # Unfortunately this setup will only work for Qt>=5.15.5.
        # Workaround for older versions is to install a codec pack.
        if os.name == 'nt':
            os.environ['QT_MULTIMEDIA_PREFERRED_PLUGINS'] = 'windowsmediafoundation'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import threading
import time
from typing import Optional

from PyQt5.QtWidgets import QStyle, QWidget

from kast.interface.qt5.UiServices import UiServices
from kast.interface.qt5.utils.dialogs import DialogQuestion
from kast.interface.qt5.service.UiEvent import UiEvent, UiState
from kast.interface.qt5.view.StatusBarView import Ui_StatusBarView
from kast.interface.qt5.viewModel.ViewModelBase import ViewBase, ViewModelBase


class View(ViewBase, QWidget, Ui_StatusBarView):
    pass


class StatusBarViewModel(ViewModelBase):

    def __init__(self, parent: QWidget, uiServices: UiServices) -> None:
        self._view = View(parent=parent)
        super().__init__(uiServices=uiServices, view=self._view)
        self.uiServices.uiEventObserver.register(self, self._onUiEvent)

        self._cancelEvent: Optional[threading.Event] = None
        self._cancelMsgBox: Optional[DialogQuestion] = None
        self._taskStartTime: Optional[float] = None

        self._view.buttonCancel.setIcon(self._view.style().standardIcon(QStyle.SP_BrowserStop))
        self._view.widgetProgress.setVisible(False)

        self._view.buttonCancel.clicked.connect(self._signalClickedCancel)

    def _onUiEvent(self, uiEvent: UiEvent) -> None:
        if uiEvent.state == UiState.Closing:
            self._cancelTask()

        self._view.labelStatusMessage.setText(self._createStatusMessage(uiEvent=uiEvent))
        self._updateProgressArea(uiEvent=uiEvent)

    def _updateProgressArea(self, uiEvent: UiEvent) -> None:
        self._view.widgetProgress.setVisible(not uiEvent.progress.complete)

        self._cancelEvent = uiEvent.progress.cancelEvent
        self._view.buttonCancel.setEnabled(self._cancelEvent is not None)

        if uiEvent.progress.complete:
            self._dismissCancelMessageBox()
            return

        vMin, vMax, vCur = (0, 100, uiEvent.progress.percentage) if uiEvent.progress.percentage is not None else (0, 0, 0)
        self._view.progressBar.setRange(vMin, vMax)
        self._view.progressBar.setValue(vCur)

    def _signalClickedCancel(self) -> None:
        try:
            self._cancelMsgBox = DialogQuestion(title="Abort", message="Do you really want to abort current background task?")
            if not self._cancelMsgBox.display():
                return

            self._cancelTask()

        finally:
            self._cancelMsgBox = None

    def _dismissCancelMessageBox(self) -> None:
        self._cancelMsgBox and self._cancelMsgBox.dismiss()

    def _cancelTask(self) -> None:
        self._cancelEvent is not None and self._cancelEvent.set()

    def _createStatusMessage(self, uiEvent: UiEvent) -> str:
        if uiEvent.state == UiState.Streaming:
            playerState = self.uiServices.uiStateService.castState.mediaState.playerState
            return f"{uiEvent.state.value} ({playerState.value.lower().capitalize()})"

        remainingEstimateMessage = self._createRemainingEstimateMessage(uiEvent=uiEvent)
        remainingEstimateMessage = f" (Remaining estimate: {remainingEstimateMessage})" if remainingEstimateMessage else ''

        return uiEvent.state.value + remainingEstimateMessage

    def _createRemainingEstimateMessage(self, uiEvent: UiEvent) -> Optional[str]:
        if uiEvent.progress.complete or not uiEvent.progress.percentage:
            self._taskStartTime = None
            return None

        if self._taskStartTime is None:
            self._taskStartTime = time.time()
            return None

        duration = time.time() - self._taskStartTime
        percentage = uiEvent.progress.percentage
        estimate = (duration / percentage) * (100 - percentage)
        return time.strftime('%H:%M:%S', time.gmtime(estimate))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import sys
import tempfile
from pathlib import Path

import psutil
from tendo import singleton


class SingletonApplication:

    def __init__(self, appName: str) -> None:
        try:
            self._appLock = singleton.SingleInstance(lockfile=self.establishLockFilePath(appName=appName))
        except singleton.SingleInstanceException:
            print(f"Application {appName} already running!")
            sys.exit(-1)

    @staticmethod
    def establishLockFilePath(appName: str) -> str:
        userName = psutil.Process().username().replace('\\', '-')
        return str(Path(tempfile.gettempdir()) / f"{appName}-{userName}-instance.lock")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from typing import Any, Callable, Dict

from kast.settings.SettingsState import SettingsState
from kast.utils.Loggable import Loggable


class SettingsEventObserver(Loggable):

    Callback = Callable[[SettingsState], None]

    def __init__(self) -> None:
        self._listeners: Dict[Any, SettingsEventObserver.Callback] = {}

    def register(self, listener: Any, callback: Callback) -> None:
        self._listeners[listener] = callback

    def unregister(self, listener: Any) -> None:
        if listener in self._listeners.keys():
            self._listeners.pop(listener)

    def notify(self, event: SettingsState) -> None:
        for callback in self._listeners.values():
            callback(event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from PyQt5.QtGui import QWindow

from kast.Services import Services
from kast.interface.qt5.UiServices import UiServices


class TaskbarViewModelBase:

    def __init__(
        self,
        parent: QWindow,
        uiServices: UiServices
    ) -> None:
        self._parent = parent
        self._uiServices = uiServices

    @property
    def services(self) -> Services:
        return self._uiServices.services

    @property
    def uiServices(self) -> UiServices:
        return self._uiServices



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import logging
import logging.handlers
import os
import platform
import signal
import sys
import threading
import types
from typing import Callable

import setproctitle

from kast.utils.Loggable import Loggable

if os.name == 'posix':
    from kast.interface.qt5.KastQtApp import KastQtApp  # Must be imported before services (to avoid media player problem)!
    from kast.Services import Services

if os.name == 'nt':
    from kast.Services import Services  # Must be imported before Qt (to avoid WinRT problem)!
    from kast.interface.qt5.KastQtApp import KastQtApp


class KastApp(Loggable):

    def __init__(self, debug: bool = False) -> None:
        self._debug = debug
        self._services = Services()
        self._desktopApp = KastQtApp(services=self._services)

    def run(self) -> int:
        setproctitle.setproctitle(self._services.settings.appName)

        self._initLogger()

        signal.signal(signal.SIGTERM, self._onSignal)
        signal.signal(signal.SIGINT, self._onSignal)

        try:
            self.log.info(f"Platform: {platform.platform()}")
            self.log.info(f"Python: {sys.version}")
            self.log.info(f"Interpreter path: '{sys.executable}'")
            self.log.info(f"Application: {self._services.settings.appName} ({self._services.settings.appVersion})")
            self.log.info(f"Persistent storage path: '{self._services.settings.persistentStoragePath}'")
            self.log.info(f"Temporary storage path: '{self._services.settings.temporaryStoragePath}'")

            self._services.mediaServer.mediaContent.thumbnailFile = self._services.settings.appIconPath
            self._services.mediaServer.start()
            self._runAsThread(self._services.osMediaOverlay.run)
            return self._desktopApp.run()

        except Exception as ex:
            self.log.exception(ex)
            return 1

        finally:
            self._onExit()

    def _onSignal(self, signum: int, frame: types.FrameType) -> None:
        self.log.info(f"Caught signal: {signal.Signals(signum).name}")
        self._desktopApp.exit(1)

    def _onExit(self) -> None:
        self._services.castController.quit()
        self._services.mediaServer.stop()

    def _runAsThread(self, callback: Callable[[], None]) -> None:
        thr = threading.Thread(target=callback)
        thr.daemon = True
        thr.start()

    def _initLogger(self) -> None:
        logFormatter = logging.Formatter(
            fmt='[%(asctime)s][%(threadName)s][%(levelname)-8.8s][%(name)s]: %(message)s',
            datefmt='%d/%m/%Y %I:%M:%S %p'
        )
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.INFO)

        fileHandler = logging.handlers.RotatingFileHandler(
            filename=self._services.settings.logFilePath,
            maxBytes=(5 * 1_000_000),
            backupCount=3
        )
        fileHandler.setFormatter(logFormatter)
        rootLogger.addHandler(fileHandler)

        if self._debug:
            consoleHandler = logging.StreamHandler()
            consoleHandler.setFormatter(logFormatter)
            rootLogger.addHandler(consoleHandler)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from abc import ABC, abstractmethod
from pathlib import Path

from kast.media.processing.MediaProcessingService import MediaProcessingService
from kast.media.processing.common import StreamId


class ISubtitlesSource(ABC):

    @abstractmethod
    def toVtt(self) -> Path:
        """Convert subtitles source to VTT format and provide result file path."""


class SubtitlesFromStream(ISubtitlesSource):

    def __init__(
        self,
        mediaProcessingService: MediaProcessingService,
        mediaFile: Path,
        streamId: StreamId
    ) -> None:
        self._mediaProcessingService = mediaProcessingService
        self._mediaFile = mediaFile
        self._streamId = streamId

    def toVtt(self) -> Path:
        return self._mediaProcessingService.extractSubtitles(
            inputFile=self._mediaFile,
            streamId=self._streamId
        )


class SubtitlesFromFile(ISubtitlesSource):

    def __init__(
        self,
        mediaProcessingService: MediaProcessingService,
        subtitlesFile: Path
    ) -> None:
        self._mediaProcessingService = mediaProcessingService
        self._subtitlesFile = subtitlesFile

    def toVtt(self) -> Path:
        return self._mediaProcessingService.convertSubtitles(inputFile=self._subtitlesFile)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J.

import socketserver
import wsgiref.simple_server as wsgi_server
from typing import Optional

import bottle

from kast.utils.Loggable import Loggable


class ServerAdapter(Loggable, bottle.ServerAdapter):

    def __init__(self, host: str, port: int) -> None:
        super().__init__(host=host, port=port)
        self._server: Optional[wsgi_server.WSGIServer] = None

    def run(self, handler: bottle.Bottle) -> None:
        class Server(socketserver.ThreadingMixIn, wsgi_server.WSGIServer):
            pass

        class RequestHandler(Loggable, wsgi_server.WSGIRequestHandler):

            def log_error(self, format, *args) -> None:
                self.log.error(format % args)

            def log_message(self, format, *args) -> None:
                self.log.info(format % args)

        self._server = wsgi_server.make_server(
            host=self.host,
            port=self.port,
            app=handler,
            server_class=Server,
            handler_class=RequestHandler
        )
        self._server.serve_forever()

    def stop(self) -> None:
        self._server.shutdown()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import os
from dataclasses import dataclass, field
from pathlib import Path

from dataclasses_json import dataclass_json


def defaultBrowseMediaPath() -> str:
    return str(Path.home() if os.name == 'posix' else Path.home().anchor)


@dataclass_json
@dataclass
class SettingsState:
    previousTemporaryPath: str = ''
    browseMediaPath: str = field(default_factory=defaultBrowseMediaPath)  # Path is not serializable.



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import argparse
import sys

from kast.KastApp import KastApp


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--debug", help="enable debug mode", action="store_true")
    parsedArgs = parser.parse_args()

    app = KastApp(debug=parsedArgs.debug)
    sys.exit(app.run())


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from pathlib import Path

StreamId = int
CodecName = str


def containerExtension(filePath: Path) -> str:
    return filePath.suffix.split('.')[-1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J.

import contextlib
import socket
import threading
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import bottle

from kast.media.streaming.ServerAdapter import ServerAdapter
from kast.utils.Loggable import Loggable


@dataclass
class MediaContent:
    movieFile: Optional[Path] = None
    subtitlesFile: Optional[Path] = None
    thumbnailFile: Optional[Path] = None


class MediaServer(Loggable):

    URL_MEDIA = '/media'
    URL_MOVIE = URL_MEDIA + '/movie'
    URL_SUBTITLES = URL_MEDIA + '/subtitles'
    URL_THUMBNAIL = URL_MEDIA + '/thumbnail'

    def __init__(self) -> None:
        self._host = self._getHost()
        self._port = self._getPort()

        self._webApp = bottle.Bottle()
        self._server = ServerAdapter(host=self._host, port=self._port)

        self._thread: Optional[threading.Thread] = None

        self._mediaContent = MediaContent()

    def start(self) -> None:
        self._thread = threading.Thread(target=self._run)
        self._thread.daemon = True
        self._thread.start()

    def stop(self) -> None:
        self._server.stop()
        self._thread.join()

    @property
    def mediaContent(self) -> MediaContent:
        return self._mediaContent

    @property
    def movieUrl(self) -> Optional[str]:
        movieFile = self.mediaContent.movieFile
        return None if movieFile is None else self._getUrl(
            urlBase=self.URL_MOVIE,
            fileName='movie',
            extension=self._getExtWithoutComa(movieFile)
            )

    @property
    def subtitleUrl(self) -> Optional[str]:
        subtitlesFile = self.mediaContent.subtitlesFile
        return None if subtitlesFile is None else self._getUrl(
            urlBase=self.URL_SUBTITLES,
            fileName='subtitles',
            extension=self._getExtWithoutComa(subtitlesFile)
        )

    @property
    def thumbnailUrl(self) -> Optional[str]:
        thumbnailFile = self.mediaContent.thumbnailFile
        return None if thumbnailFile is None else self._getUrl(
            urlBase=self.URL_THUMBNAIL,
            fileName='thumbnail',
            extension=self._getExtWithoutComa(thumbnailFile)
        )

    def _run(self) -> None:
        web = self._webApp

        @web.get(self.URL_MOVIE + '/<name>.<ext>')
        def movie(name: str, ext: str) -> bottle.Response:
            return self._serveFile(self.mediaContent.movieFile)

        @web.get(self.URL_SUBTITLES + '/<name>.<ext>')
        def subtitles(name: str, ext: str) -> bottle.Response:
            return self._serveFile(self.mediaContent.subtitlesFile)

        @web.get(self.URL_THUMBNAIL + '/<name>.<ext>')
        def thumbnail(name: str, ext: str) -> bottle.Response:
            return self._serveFile(self.mediaContent.thumbnailFile)

        self._webApp.run(server=self._server, quiet=True)

    def _serveFile(self, filePath: Path) -> bottle.Response:
        if not filePath:
            return bottle.HTTPError(410, 'Resource no longer available.')
        response = bottle.static_file(filename=filePath.name, root=filePath.parent)

        if 'Last-Modified' in response.headers:
            del response.headers['Last-Modified']

        response.headers['Access-Control-Allow-Origin'] = '*'
        response.headers['Access-Control-Allow-Methods'] = 'GET, HEAD'
        response.headers['Access-Control-Allow-Headers'] = 'Content-Type'

        return response

    def _getUrl(self, urlBase: str, fileName: str, extension: str) -> str:
        return f"http://{self._host}:{self._port}{urlBase}/{fileName}.{extension}"

    @staticmethod
    def _getExtWithoutComa(filePath: Path) -> str:
        return filePath.suffix.split('.', maxsplit=1)[-1]

    @staticmethod
    def _getHost() -> str:
        hostIps = socket.gethostbyname_ex(socket.gethostname())[2]
        hostIps = [ip for ip in hostIps if not ip.startswith("127.")]
        if hostIps:
            return hostIps[0]

        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.connect(("8.8.8.8", 53))
            return sock.getsockname()[0]

    @staticmethod
    def _getPort() -> int:
        with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(('0.0.0.0', 0))
            return s.getsockname()[1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import threading
from dataclasses import dataclass
from enum import Enum
from typing import Optional


class UiState(Enum):
    Idle = 'Idle'
    DeviceSearch = 'Searching for cast devices'
    VideoProbe = 'Extracting video meta data'
    Preprocessing = 'Preprocessing'
    AvProcessing = 'Transcoding/Remuxing'
    Connecting = 'Connecting'
    Streaming = 'Streaming'
    Closing = 'Closing'


@dataclass
class Progress:
    complete: bool = True
    percentage: Optional[int] = None
    cancelEvent: Optional[threading.Event] = None


@dataclass
class UiEvent:
    state: UiState = UiState.Idle
    progress: Progress = Progress()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

import argparse
import sys
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum, auto
from pathlib import Path
from typing import Dict, Optional, Tuple, Any, List

DIR_ROOT = Path(__file__).resolve().parent


def makeReqFilePath(suffix: Optional[str] = None) -> Path:
    suffix = f'_{suffix}' if suffix else ''
    return DIR_ROOT / f'requirements{suffix}.txt'


@dataclass
class Package:
    name: str
    version: Optional[str] = None
    osName: Optional[str] = None


Packages = Dict[str, Package]


def pkg(*args, **kwargs) -> Tuple[str, Package]:
    package = Package(*args, **kwargs)
    return package.name, package


@dataclass
class Requirements:
    filePath: Path
    packages: Packages


requirementsPkg = Requirements(
    filePath=makeReqFilePath(),
    packages=OrderedDict([
        pkg('typing_extensions', '3.10.0', 'posix'),  # Exists because of MPRIS Server.
        pkg('mpris_server', '0.3.4', 'posix'),

        pkg('winrt', '1.0.21033.1', 'nt'),

        pkg('tendo', '0.2.15'),
        pkg('psutil', '5.9.0'),
        pkg('av', '9.2.0'),
        pkg('pysubs2', '1.4.2'),
        pkg('pycaption', '2.0.8'),
        pkg('bottle', '0.12.19'),
        pkg('zeroconf', '0.30.0'),  # Exists because of PyChromecast.
        pkg('pychromecast', '11.0.0'),
        pkg('dataclasses-json', '0.5.7'),
        pkg('setproctitle', '1.2.3'),
        pkg('pyqt5', '5.15.6'),
    ]))

requirementsDev = Requirements(
    filePath=makeReqFilePath('dev'),
    packages=OrderedDict([
        pkg('setuptools'),
        pkg('setup-qt', '1.0.0'),
    ]))

requirementsDist = Requirements(
    filePath=makeReqFilePath('dist'),
    packages=OrderedDict([
        pkg('twine'),
        pkg('PyInstaller'),

        pkg('Pillow', osName='nt'),
    ]))


class EnumBase(Enum):

    def __str__(self):
        return self.name.lower()

    def __call__(self, *args, **kwargs):
        return str(self)

    @classmethod
    def all(cls):
        return [target() for target in cls]


class Arg(EnumBase):
    Target = 'use stable (default) dependencies or try latest'
    Pkg = 'create package requirements file'
    Dev = 'create development requirements file'
    Dist = 'create distribution requirements file'

    def help(self):
        return self.value


class Target(EnumBase):
    Stable = auto()
    Latest = auto()


REQUIREMENTS = {
    Arg.Pkg(): requirementsPkg,
    Arg.Dev(): requirementsDev,
    Arg.Dist(): requirementsDist,
}


def generate(*args: Any) -> None:
    parser = argparse.ArgumentParser()

    parser.add_argument(Arg.Target(), help=Arg.Target.help(), choices=Target.all(), nargs='?', default=Target.Stable())
    parser.add_argument(f'--{Arg.Pkg()}', help=Arg.Pkg.help(), action="store_true")
    parser.add_argument(f'--{Arg.Dev()}', help=Arg.Dev.help(), action="store_true")
    parser.add_argument(f'--{Arg.Dist()}', help=Arg.Dist.help(), action="store_true")

    parsedArgs = parser.parse_args(args)

    parsedArgsDict = vars(parsedArgs)
    genRegRequests = [parsedArgsDict[reqName] for reqName in REQUIREMENTS.keys()]
    if not any(genRegRequests):
        print("<!> No requirements files were selected!\n(Will fall back to generate package requirements file.)")
        parsedArgsDict[Arg.Pkg()] = True

    target = parsedArgsDict[Arg.Target()]
    separator = '==' if target == Target.Stable() else '>='

    print(f" -> Dependencies versions: {target}")

    def packageToLine(package: Package) -> str:
        return f'{package.name}' \
            + (f' {separator} {package.version}' if package.version else '') \
            + (f'; os_name == "{package.osName}"' if package.osName else '') \
            + '\n'

    def packgesToLines(packages: Packages) -> List[str]:
        return [packageToLine(package) for package in packages.values()]

    def linesToFile(filePath: Path, lines: List[str]) -> None:
        print(f" -> Generating file: '{filePath.name}'")
        with open(filePath, 'w') as fOutput:
            fOutput.writelines(lines)

    for reqName, requirements in REQUIREMENTS.items():
        if not parsedArgsDict[reqName]:
            continue

        linesToFile(requirements.filePath, packgesToLines(requirements.packages))


if __name__ == "__main__":
    generate(*sys.argv[1:])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from PyQt5.QtWidgets import QMessageBox

ButtonId = int


class DialogQuestion:

    def __init__(
        self,
        title: str,
        message: str,
        default: bool = False,
        icon: QMessageBox.Icon = QMessageBox.Question,
        positiveButton: ButtonId = QMessageBox.Ok,
        negativeButton: ButtonId = QMessageBox.Cancel
    ) -> None:
        self._positiveButton = positiveButton
        self._negativeButton = negativeButton
        self._msgBox = msgBox = QMessageBox()

        msgBox.setIcon(icon)
        msgBox.setWindowTitle(title)
        msgBox.setText(message)
        msgBox.setStandardButtons(positiveButton | negativeButton)
        msgBox.setDefaultButton(positiveButton if default else negativeButton)

    def display(self) -> bool:
        return self._msgBox.exec() == self._positiveButton

    def dismiss(self) -> None:
        self._msgBox.close()


def dialogCriticalError(message: str) -> None:
    _dialogInformation(title="CriticalError", message=message, icon=QMessageBox.Critical)


def dialogError(message: str) -> None:
    _dialogInformation(title="Error", message=message, icon=QMessageBox.Critical)


def dialogWarning(message: str) -> None:
    _dialogInformation(title="Warning", message=message, icon=QMessageBox.Warning)


def dialogQuestionOkCancel(title: str, message: str, default: bool = False) -> bool:
    return DialogQuestion(
        title=title,
        message=message,
        default=default
    ).display()


def dialogQuestionYesNo(title: str, message: str, default: bool = False) -> bool:
    return DialogQuestion(
        title=title,
        message=message,
        default=default,
        positiveButton=QMessageBox.Yes,
        negativeButton=QMessageBox.No
    ).display()


def _dialogInformation(title: str, message: str, icon: QMessageBox.Icon = QMessageBox.Information) -> None:
    msgBox = QMessageBox()
    msgBox.setIcon(icon)
    msgBox.setWindowTitle(title)
    msgBox.setText(message)
    msgBox.exec()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright by: P.J. Grochowski

from pathlib import Path
from typing import List, Optional

from PyQt5.QtWidgets import QComboBox, QFileDialog, QStyle, QWidget

from kast.interface.qt5.UiServices import UiServices
from kast.interface.qt5.utils.dialogs import dialogWarning
from kast.interface.qt5.service.UiEvent import Progress, UiEvent, UiState
from kast.interface.qt5.view.VideoSettingsView import Ui_VideoSettingsView
from kast.interface.qt5.viewModel.ViewModelBase import ViewBase, ViewModelBase
from kast.media.processing.MetaData import MetaData
from kast.media.processing.SubtitlesSourceList import SubtitlesSourceList
from kast.media.processing.Transcoder import TranscodeParams
from kast.utils.Loggable import Loggable


class View(ViewBase, QWidget, Ui_VideoSettingsView):
    pass


class VideoSettingsViewModel(ViewModelBase, Loggable):

    def __init__(self, parent: QWidget, uiServices: UiServices) -> None:
        self._view = View(parent=parent)
        super().__init__(uiServices=uiServices, view=self._view)
        self.uiServices.uiEventObserver.register(self, self._onUiEvent)

        self._view.buttonDeviceRefresh.setIcon(self._view.style().standardIcon(QStyle.SP_BrowserReload))
        self._view.buttonVideoOpen.setIcon(self._view.style().standardIcon(QStyle.SP_DirOpenIcon))
        self._view.buttonSubtitlesAdd.setIcon(self._view.style().standardIcon(QStyle.SP_DirOpenIcon))

        self._view.buttonDeviceRefresh.clicked.connect(self._signalDevicesRefresh)
        self._view.buttonVideoOpen.clicked.connect(self._signalVideoOpen)
        self._view.buttonSubtitlesAdd.clicked.connect(self._signalSubtitlesAdd)
        self._view.buttonStream.clicked.connect(self._signalStream)

        self._metaData = MetaData()
        self._subtitlesSourceList = SubtitlesSourceList()
        self._lastTranscodeParams: Optional[TranscodeParams] = None

        self._fillSubtitlesComboBox()

    def _onUiEvent(self, uiEvent: UiEvent) -> None:
        viewEnabled = uiEvent.state == UiState.Idle
        self._view.setEnabled(viewEnabled)

        self._view.setVisible(uiEvent.state != UiState.Streaming)

    def _signalDevicesRefresh(self) -> None:
        self.uiServices.uiStateService.dispatch(UiEvent(state=UiState.DeviceSearch, progress=Progress(complete=False)))

        def interfaceCallback(deviceNames: List[str]) -> None:
            self._view.comboBoxDevice.clear()
            self._view.comboBoxDevice.addItems(deviceNames)

            if not deviceNames:
                message = "No devices could be found!"
                "\n(Make sure this PC and your cast device are in the same network.)"
                dialogWarning(message=message)

            self.uiServices.uiStateService.dispatch(UiEvent(state=UiState.Idle))

        def backgroundTask() -> None:
            devicesList = self.services.castController.searchDevices()
            self.uiServices.interfaceScheduler.schedule(lambda: interfaceCallback(deviceNames=devicesList))

        self.uiServices.backgroundRunner.execute(lambda: backgroundTask())

    def _signalVideoOpen(self) -> None:
        filePath = QFileDialog.getOpenFileName(
            self._view,
            "Open Video",
            str(self.services.settings.browseMediaDir),
            "Videos (*.mp4 *.mkv *.webm *.avi)"
        )[0]
        if not filePath:
            return

        self._view.lineEditVideo.setText(filePath)

        def videoOpenedCallback(metaData: MetaData) -> None:
            self._fillAudioComboBox(items=metaData.audioStreamLangs)
            self._fillSubtitlesComboBox(items=metaData.subtitleStreamLangs)

        self.uiServices.mediaControlService.openVideo(Path(filePath), videoOpenedCallback)

    def _signalSubtitlesAdd(self) -> None:
        filePath = QFileDialog.getOpenFileName(
            self._view,
            "Open Subtitles",
            str(self.services.settings.browseMediaDir),
            "Subtitles (*.srt *.sub *.ass *.ssa *.txt *.vtt)"
        )[0]
        if not filePath:
            return

        filePath = Path(filePath)

        self.uiServices.mediaControlService.addSubtitles(filePath)

        self._view.comboBoxSubtitles.addItem(filePath.name)
        self._view.comboBoxSubtitles.setCurrentIndex(self._view.comboBoxSubtitles.count() - 1)

    def _signalStream(self) -> None:
        errors = []

        deviceName = self._view.comboBoxDevice.currentText()
        (not deviceName) and errors.append("No cast device has been selected!")

        videoFilePath = str(self._view.lineEditVideo.text())
        (not videoFilePath) and errors.append("No video file has been selected!")

        if errors:
            errors = ["Could not start streaming! Preconditions that failed:"] + errors
            dialogWarning(message='\n - '.join(errors))
            return

        self.uiServices.mediaControlService.startStream(
            deviceName=deviceName,
            audioStreamId=self._getSelectedAudioStreamId(),
            subtitlesId=self._getSelectedSubtitlesId()
        )

    def _getSelectedAudioStreamId(self) -> int:
        return self._view.comboBoxAudio.currentIndex()  # TODO: Should we care if it fails (-1)?

    def _getSelectedSubtitlesId(self) -> Optional[int]:
        selectedSubtitles = self._view.comboBoxSubtitles.currentIndex()
        return (selectedSubtitles - 1) if selectedSubtitles > 0 else None

    def _fillDeviceComboBox(self, items: List[str]) -> None:
        self._fillComboBox(self._view.comboBoxDevice, items)

    def _fillAudioComboBox(self, items: List[str] = None) -> None:
        self._fillComboBox(self._view.comboBoxAudio, items)

    def _fillSubtitlesComboBox(self, items: List[str] = None) -> None:
        items = items if items else []
        self._fillComboBox(self._view.comboBoxSubtitles, ['No Subtitles'] + items)

    @staticmethod
    def _fillComboBox(comboBox: QComboBox, items: List[str] = None) -> None:
        comboBox.clear()
        items and comboBox.addItems(items)
